# !/usr/bin/python
# coding=utf-8
from PySide2 import QtWidgets, QtCore

from uitk.widgets.attributes import Attributes
from uitk.widgets.text import RichText, TextOverlay
from uitk.widgets.menu import MenuInstance


class CheckBox(QtWidgets.QCheckBox, MenuInstance, Attributes, RichText, TextOverlay):
	'''
	'''
	def __init__(self, parent=None, **kwargs):
		super().__init__(parent)

		self.setStyleSheet(parent.styleSheet()) if parent else None

		self.setCheckBoxRichTextStyle(self.isChecked()) #set the initial style for rich text depending on the current state.
		self.stateChanged.connect(lambda state: self.setCheckBoxRichTextStyle(state)) #set the style on future state changes.

		#override built-ins
		self.text = self.richText
		self.setText = self.setRichText
		self.sizeHint = self.richTextSizeHint

		self.setAttributes(**kwargs)


	def setCheckBoxRichTextStyle(self, state):
		'''
		'''
		if self.hasRichText:
			self.setRichTextStyle(textColor='black' if state>0 else 'white')


	def checkState_(self):
		'''Get the state of a checkbox as an integer value.
		Simplifies working with tri-state checkboxes.
		'''
		if self.isTristate():		
			state = {QtCore.Qt.CheckState.Unchecked:0, QtCore.Qt.CheckState.PartiallyChecked:1, QtCore.Qt.CheckState.Checked:2}
			return state[self.checkState()]

		else:
			return 1 if self.isChecked() else 0


	def setCheckState_(self, state):
		'''Set the state of a checkbox as an integer value.
		Simplifies working with tri-state checkboxes.

		Parameters:
			state (int)(bool): 0 or False: unchecked, 1 or True: checked. 
				If tri-state: 0: unchecked, 1: paritally checked, 2: checked.
		'''
		if self.isTristate():		
			s = {0:QtCore.Qt.CheckState.Unchecked, 1:QtCore.Qt.CheckState.PartiallyChecked, 2:QtCore.Qt.CheckState.Checked}
			return self.setCheckState(s[state])

		else:
			self.setChecked(state)


	def mousePressEvent(self, event):
		'''
		Parameters:
			event (QEvent)

		Return:
			(QEvent)
		'''
		if event.button()==QtCore.Qt.RightButton:
			if self.ctxMenu:
				self.ctxMenu.show()

		super().mousePressEvent(event)


	def showEvent(self, event):
		'''
		Parameters:
			event=<QEvent>
		'''
		# self.setTextOverlay('±', alignment='AlignRight')

		super().showEvent(event)









if __name__ == "__main__":
	import sys
	from PySide2.QtCore import QSize
	app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv) #return the existing QApplication object, or create a new one if none exists.

	w = CheckBox(
		parent=None,
		setObjectName='chk000',
		setText='A Check Box <b>w/Rich Text</b>',
		resize=QSize(125, 45),
		setWhatsThis='',
		setChecked=False,
		setVisible=True,
	)
	# w.show()

	sys.exit(app.exec_())



# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

'''
Promoting a widget in designer to use a custom class:
>	In Qt Designer, select all the widgets you want to replace, 
		then right-click them and select 'Promote to...'. 

>	In the dialog:
		Base Class:		Class from which you inherit. ie. QWidget
		Promoted Class:	Name of the class. ie. "MyWidget"
		Header File:	Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>	Then click "Add", "Promote", 
		and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
'''

# Deprecated:




# !/usr/bin/python
# coding=utf-8
from uitk.slots import Slots



class Materials(Slots):
	'''
	'''
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		'''
		'''
		self.sb.materials_submenu.b003.setVisible(False) #hide the current material submenu button until a material is available.

		cmb002 = self.sb.materials.cmb002
		cmb002.ctxMenu.add('QComboBox', setObjectName='cmb001', addItems=['Scene Materials', 'ID Map Materials', 'Favorite Materials'], setToolTip='Filter materials list based on type.')
		cmb002.ctxMenu.add(self.sb.Label, setText='Open in Editor', setObjectName='lbl000', setToolTip='Open material in editor.')
		cmb002.ctxMenu.add(self.sb.Label, setText='Rename', setObjectName='lbl001', setToolTip='Rename the current material.')
		cmb002.ctxMenu.add(self.sb.Label, setText='Delete', setObjectName='lbl002', setToolTip='Delete the current material.')
		cmb002.ctxMenu.add(self.sb.Label, setText='Delete All Unused Materials', setObjectName='lbl003', setToolTip='Delete All unused materials.')
		cmb002.beforePopupShown.connect(self.cmb002) #refresh comboBox contents before showing it's menu.
		cmb002.returnPressed.connect(lambda: self.lbl001(setEditable=False))
		cmb002.currentIndexChanged.connect(lambda: cmb002.ctxMenu.setTitle(cmb002.currentText())) #set the popup title to be the current materials name.
		cmb002.ctxMenu.cmb001.currentIndexChanged.connect(lambda: self.sb.materials.group000.setTitle(cmb002.ctxMenu.cmb001.currentText())) #set the groupbox title to reflect the current filter.
		cmb002.ctxMenu.cmb001.currentIndexChanged.connect(self.cmb002) #refresh cmb002 contents.
		self.cmb002() #initialize the materials list

		tb000 = self.sb.materials.tb000
		tb000.ctxMenu.add('QCheckBox', setText='All Objects', setObjectName='chk003', setToolTip='Search all scene objects, or only those currently selected.')
		tb000.ctxMenu.add('QCheckBox', setText='Shell', setObjectName='chk005', setToolTip='Select entire shell.')
		tb000.ctxMenu.add('QCheckBox', setText='Invert', setObjectName='chk006', setToolTip='Invert Selection.')

		tb002 = self.sb.materials.tb002
		tb002.ctxMenu.add('QRadioButton', setText='Current Material', setObjectName='chk007', setChecked=True, setToolTip='Re-Assign the current stored material.')
		tb002.ctxMenu.add('QRadioButton', setText='New Material', setObjectName='chk009', setToolTip='Assign a new material.')
		tb002.ctxMenu.add('QRadioButton', setText='New Random Material', setObjectName='chk008', setToolTip='Assign a new random ID material.')
		tb002.ctxMenu.chk007.clicked.connect(lambda state: tb002.setText('Assign Current')) 
		tb002.ctxMenu.chk009.clicked.connect(lambda state: tb002.setText('Assign New'))
		tb002.ctxMenu.chk008.clicked.connect(lambda state: tb002.setText('Assign Random'))


	def draggable_header(self, state=None):
		'''Context menu
		'''
		dh = self.sb.materials.draggable_header


	def b000(self):
		'''Material List: Delete
		'''
		self.lbl002()


	def b001(self):
		'''Material List: Edit
		'''
		self.lbl000()


	def b003(self):
		'''Assign: Assign Current
		'''
		self.sb.materials.tb002.ctxMenu.chk007.setChecked(True)
		self.sb.materials.tb002.setText('Assign Current')
		self.tb002()


	def b004(self):
		'''Assign: Assign Random
		'''
		self.sb.materials.tb002.ctxMenu.chk008.setChecked(True)
		self.sb.materials.tb002.setText('Assign Random')
		self.tb002()


	def b005(self):
		'''Assign: Assign New
		'''
		self.sb.materials.tb002.ctxMenu.chk009.setChecked(True)
		self.sb.materials.tb002.setText('Assign New')
		self.tb002()




# !/usr/bin/python
# coding=utf-8
from uitk.slots.maya import *
from uitk.slots.selection import Selection



class Selection_maya(Selection, Slots_maya):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		dh = self.sb.selection.draggable_header
		items = ['Polygon Selection Constraints']
		dh.ctxMenu.cmb000.addItems_(items, 'Selection Editors:')

		cmb002 = self.sb.selection.cmb002
		items = ['IK Handles','Joints','Clusters','Lattices','Sculpt Objects','Wires','Transforms','Geometry','NURBS Curves','NURBS Surfaces','Polygon Geometry','Cameras','Lights','Image Planes','Assets','Fluids','Particles','Rigid Bodies','Rigid Constraints','Brushes','Strokes','Dynamic Constraints','Follicles','nCloths','nParticles','nRigids']
		cmb002.addItems_(items, 'By Type:')

		cmb003 = self.sb.selection.cmb003
		items = ['Verts', 'Vertex Faces', 'Vertex Perimeter', 'Edges', 'Edge Loop', 'Edge Ring', 'Contained Edges', 'Edge Perimeter', 'Border Edges', 'Faces', 'Face Path', 'Contained Faces', 'Face Perimeter', 'UV\'s', 'UV Shell', 'UV Shell Border', 'UV Perimeter', 'UV Edge Loop', 'Shell', 'Shell Border'] 
		cmb003.addItems_(items, 'Convert To:')

		cmb005 = self.sb.selection.cmb005
		items = ['Angle', 'Border', 'Edge Loop', 'Edge Ring', 'Shell', 'UV Edge Loop']
		cmb005.addItems_(items, 'Off')


	def txt001(self):
		'''Select By Name
		'''
		searchStr = str(self.sb.selection.txt001.text()) #asterisk denotes startswith*, *endswith, *contains* 
		if searchStr:
			selection = pm.select(pm.ls (searchStr))


	def lbl000(self):
		'''Selection Sets: Create New
		'''
		cmb = self.sb.selection.cmb001
		if not cmb.isEditable():
			cmb.addItems_('', ascending=True)
			cmb.setEditable(True)
			cmb.lineEdit().setPlaceholderText('New Set:')
		else:
			name = cmb.currentText()
			self.creatNewSelectionSet(name)
			self.cmb001() #refresh the sets comboBox
			cmb.setCurrentIndex(0)


	def lbl001(self):
		'''Selection Sets: Modify Current
		'''
		cmb = self.sb.selection.cmb001
		if not cmb.isEditable():
			name = cmb.currentText()
			self._oldSetName = name
			cmb.setEditable(True)
			cmb.lineEdit().setPlaceholderText(name)
		else:
			name = cmb.currentText()
			self.modifySet(self._oldSetName)
			cmb.setItemText(cmb.currentIndex(), name)
			# self.cmb001() #refresh the sets comboBox


	def lbl002(self):
		'''Selection Sets: Delete Current
		'''
		cmb = self.sb.selection.cmb001
		name = cmb.currentText()

		pm.delete(name)

		self.cmb001() #refresh the sets comboBox


	def lbl003(self):
		'''Grow Selection
		'''
		pm.mel.GrowPolygonSelectionRegion()


	def lbl004(self):
		'''Shrink Selection
		'''
		pm.mel.ShrinkPolygonSelectionRegion()


	def lbl005(self):
		'''Selection Sets: Select Current
		'''
		cmb = self.sb.selection.cmb001
		name = cmb.currentText()

		if cmb.currentIndex()>0:
			pm.select(name) # pm.select(name, noExpand=1) #Select The Selection Set Itself (Not Members Of) (noExpand=select set)


	@Slots.hideMain
	def chk004(self, state=None):
		'''Ignore Backfacing (Camera Based Selection)
		'''
		if self.selection_submenu_ui.chk004.isChecked():
			pm.selectPref(useDepth=True)
			self.sb.messageBox('Camera-based selection <hl>On</hl>.', messageType='Result')
		else:
			pm.selectPref(useDepth=False)
			self.sb.messageBox('Camera-based selection <hl>Off</hl>.', messageType='Result')


	def chk008(self, state=None):
		'''Toggle Soft Selection
		'''
		if self.selection_submenu_ui.chk008.isChecked():
			pm.softSelect(edit=1, softSelectEnabled=True)
			self.sb.messageBox('Soft Select <hl>On</hl>.', messageType='Result')
		else:
			pm.softSelect(edit=1, softSelectEnabled=False)
			self.sb.messageBox('Soft Select <hl>Off</hl>.', messageType='Result')


	def cmb000(self, index=-1):
		'''Editors
		'''
		cmb = self.sb.selection.draggable_header.ctxMenu.cmb000

		if index>0:
			text = cmb.items[index]
			if text=='Polygon Selection Constraints':
				pm.mel.PolygonSelectionConstraints()
			cmb.setCurrentIndex(0)


	def cmb001(self, index=-1):
		'''Selection Sets
		'''
		cmb = self.sb.selection.cmb001

		items = [str(s) for s in pm.ls(et='objectSet', flatten=1)]
		cmb.addItems_(items, clear=True)


	def cmb002(self, index=-1):
		'''Select by Type
		'''
		cmb = self.sb.selection.cmb002

		if index>0:
			text = cmb.items[index]
			if text=='IK Handles': #
				type_ = pm.ls(type=['ikHandle', 'hikEffector'])
			elif text=='Joints': #
				type_ = pm.ls(type='joint')
			elif text=='Clusters': #
				type_ = pm.listTransforms(type='clusterHandle')
			elif text=='Lattices': #
				type_ = pm.listTransforms(type='lattice')
			elif text=='Sculpt Objects': #
				type_ = pm.listTransforms(type=['implicitSphere', 'sculpt'])
			elif text=='Wires': #
				type_ = pm.ls(type='wire')
			elif text=='Transforms': #
				type_ = pm.ls(type='transform')
			elif text=='Geometry': #Select all Geometry
				geometry = pm.ls(geometry=True)
				type_ = pm.listRelatives(geometry, p=True, path=True) #pm.listTransforms(type='nRigid')
			elif text=='NURBS Curves': #
				type_ = pm.listTransforms(type='nurbsCurve')
			elif text=='NURBS Surfaces': #
				type_ = pm.ls(type='nurbsSurface')
			elif text=='Polygon Geometry': #
				type_ = pm.listTransforms(type='mesh')
			elif text=='Cameras': #
				type_ = pm.listTransforms(cameras=1)
			elif text=='Lights': #
				type_ = pm.listTransforms(lights=1)
			elif text=='Image Planes': #
				type_ = pm.ls(type='imagePlane')
			elif text=='Assets': #
				type_ = pm.ls(type=['container', 'dagContainer'])
			elif text=='Fluids': #
				type_ = pm.listTransforms(type='fluidShape')
			elif text=='Particles': #
				type_ = pm.listTransforms(type='particle')
			elif text=='Rigid Bodies': #
				type_ = pm.listTransforms(type='rigidBody')
			elif text=='Rigid Constraints': #
				type_ = pm.ls(type='rigidConstraint')
			elif text=='Brushes': #
				type_ = pm.ls(type='brush')
			elif text=='Strokes': #
				type_ = pm.listTransforms(type='stroke')
			elif text=='Dynamic Constraints': #
				type_ = pm.listTransforms(type='dynamicConstraint')
			elif text=='Follicles': #
				type_ = pm.listTransforms(type='follicle')
			elif text=='nCloths': #
				type_ = pm.listTransforms(type='nCloth')
			elif text=='nParticles': #
				type_ = pm.listTransforms(type='nParticle')
			elif text=='nRigids': #
				type_ = pm.listTransforms(type='nRigid')

			pm.select(type_)
			cmb.setCurrentIndex(0)


	def cmb003(self, index=-1):
		'''Convert To
		'''
		cmb = self.sb.selection.cmb003

		if index>0:
			text = cmb.items[index]
			if text=='Verts': #Convert Selection To Vertices
				pm.mel.PolySelectConvert(3)
			elif text=='Vertex Faces': #
				pm.mel.PolySelectConvert(5)
			elif text=='Vertex Perimeter': #
				pm.mel.ConvertSelectionToVertexPerimeter()
			elif text=='Edges': #Convert Selection To Edges
				pm.mel.PolySelectConvert(2)
			elif text=='Edge Loop': #
				pm.mel.polySelectSp(loop=1)
			elif text=='Edge Ring': #Convert Selection To Edge Ring
				pm.mel.SelectEdgeRingSp()
			elif text=='Contained Edges': #
				pm.mel.PolySelectConvert(20)
			elif text=='Edge Perimeter': #
				pm.mel.ConvertSelectionToEdgePerimeter()
			elif text=='Border Edges': #
				pm.select(self.getBorderEdgeFromFace())
			elif text=='Faces': #Convert Selection To Faces
				pm.mel.PolySelectConvert(1)
			elif text=='Face Path': #
				pm.mel.polySelectEdges('edgeRing')
			elif text=='Contained Faces': #
				pm.mel.PolySelectConvert(10)
			elif text=='Face Perimeter': #
				pm.mel.polySelectFacePerimeter()
			elif text=='UV\'s': #
				pm.mel.PolySelectConvert(4)
			elif text=='UV Shell': #
				pm.mel.polySelectBorderShell(0)
			elif text=='UV Shell Border': #
				pm.mel.polySelectBorderShell(1)
			elif text=='UV Perimeter': #
				pm.mel.ConvertSelectionToUVPerimeter()
			elif text=='UV Edge Loop': #
				pm.mel.polySelectEdges('edgeUVLoopOrBorder')
			elif text=='Shell': #
				pm.mel.polyConvertToShell()
			elif text=='Shell Border': #
				pm.mel.polyConvertToShellBorder()
			cmb.setCurrentIndex(0)


	def cmb005(self, index=-1):
		'''Selection Contraints
		'''
		cmb = self.sb.selection.cmb005

		if index>0:
			text = cmb.items[index]
			if text=='Angle':
				pm.mel.dR_selConstraintAngle() #dR_DoCmd("selConstraintAngle");
			elif text=='Border':
				pm.mel.dR_selConstraintBorder() #dR_DoCmd("selConstraintBorder");
			elif text=='Edge Loop':
				pm.mel.dR_selConstraintEdgeLoop() #dR_DoCmd("selConstraintEdgeLoop");
			elif text=='Edge Ring':
				pm.mel.dR_selConstraintEdgeRing() #dR_DoCmd("selConstraintEdgeRing");
			elif text=='Shell':
				pm.mel.dR_selConstraintElement() #dR_DoCmd("selConstraintElement");
			elif text=='UV Edge Loop':
				pm.mel.dR_selConstraintUVEdgeLoop() #dR_DoCmd("selConstraintUVEdgeLoop");
		else:
			pm.mel.dR_selConstraintOff() #dR_DoCmd("selConstraintOff");


	def cmb006(self, index=-1):
		'''Currently Selected Objects
		'''
		cmb = self.sb.selection.draggable_header.ctxMenu.cmb006

		cmb.clear()
		items = [str(i) for i in pm.ls(sl=1, flatten=1)]
		widgets = [cmb.menu_.add('QCheckBox', setText=t, setChecked=1) for t in items[:50]] #selection list is capped with a slice at 50 elements.

		for w in widgets:
			try:
				w.disconnect() #disconnect all previous connections.

			except TypeError:
				pass #if no connections are present; pass
			w.toggled.connect(lambda state, widget=w: self.chkxxx(state=state, widget=widget))


	def chkxxx(self, **kwargs):
		'''Transform Constraints: Constraint CheckBoxes
		'''
		try:
			pm.select(kwargs['widget'].text(), deselect=(not kwargs['state']))
		except KeyError:
			pass


	def tb000(self, state=None):
		'''Select Nth
		'''
		tb = self.sb.selection.tb000

		edgeRing = tb.ctxMenu.chk000.isChecked()
		edgeLoop = tb.ctxMenu.chk001.isChecked()
		pathAlongLoop = tb.ctxMenu.chk009.isChecked()
		shortestPath = tb.ctxMenu.chk002.isChecked()
		borderEdges = tb.ctxMenu.chk010.isChecked()
		step = tb.ctxMenu.s003.value()

		selection = pm.ls(sl=1)
		if not selection:
			self.sb.messageBox('Operation requires a valid selection.')
			return

		result=[]
		if edgeRing:
			result = mtk.Cmpt.getEdgePath(selection, 'edgeRing')

		elif edgeLoop:
			result = mtk.Cmpt.getEdgePath(selection, 'edgeLoop')

		elif pathAlongLoop:
			result = mtk.Cmpt.getEdgePath(selection, 'edgeLoopPath')

		elif shortestPath:
			result = mtk.Cmpt.getShortestPath(selection)

		elif borderEdges:
			result = mtk.Cmpt.getBorderComponents(selection, 'edges')

		pm.select(result[::step])


	def tb001(self, state=None):
		'''Select Similar
		'''
		tb = self.sb.selection.tb001

		tol = tb.ctxMenu.s000.value() #tolerance
		v = tb.ctxMenu.chk011.isChecked() #vertex
		e = tb.ctxMenu.chk012.isChecked() #edge
		f = tb.ctxMenu.chk013.isChecked() #face
		t = tb.ctxMenu.chk014.isChecked() #triangle
		s = tb.ctxMenu.chk015.isChecked() #shell
		uv = tb.ctxMenu.chk016.isChecked() #uvcoord
		a = tb.ctxMenu.chk017.isChecked() #area
		wa = tb.ctxMenu.chk018.isChecked() #world area
		b = tb.ctxMenu.chk019.isChecked() #bounding box
		inc = tb.ctxMenu.chk020.isChecked() #select the original objects

		objMode = pm.selectMode(query=1, object=1)
		if objMode:
			selection = pm.ls(sl=1, objectsOnly=1, type='transform')
			pm.select(clear=1)
			for obj in selection:
				similar = self.sb.edit.slots.getSimilarMesh(obj, tol=tol, includeOrig=inc, vertex=v, edge=e, face=f, uvcoord=uv, triangle=t, shell=s, boundingBox=b, area=a, worldArea=wa)
				pm.select(similar, add=True)
		else:
			pm.mel.doSelectSimilar(1, {tol})


	def tb002(self, state=None):
		'''Select Island: Select Polygon Face Island
		'''
		tb = self.sb.selection.tb002

		rangeX = float(tb.ctxMenu.s002.value())
		rangeY = float(tb.ctxMenu.s004.value())
		rangeZ = float(tb.ctxMenu.s005.value())

		sel = pm.ls(sl=1)
		selectedFaces = mtk.Cmpt.getComponents(sel, componentType='faces')
		if not selectedFaces:
			self.sb.messageBox('The operation requires a face selection.')
			return

		similarFaces = self.sb.normals.slots.getFacesWithSimilarNormals(selectedFaces, rangeX=rangeX, rangeY=rangeY, rangeZ=rangeZ)
		islands = mtk.Cmpt.getContigiousIslands(similarFaces)
		island = [i for i in islands if bool(set(i) & set(selectedFaces))]
		pm.select(island)


	def tb003(self, state=None):
		'''Select Edges By Angle
		'''
		tb = self.sb.selection.tb003

		angleLow = tb.ctxMenu.s006.value()
		angleHigh = tb.ctxMenu.s007.value()

		objects = pm.ls(sl=1, objectsOnly=1)
		edges = Slots_maya.getEdgesByNormalAngle(objects, lowAngle=angleLow, highAngle=angleHigh)
		pm.select(edges)

		pm.selectMode(component=1)
		pm.selectType(edge=1)


	def b016(self):
		'''Convert Selection To Vertices
		'''
		pm.mel.PolySelectConvert(3)


	def b017(self):
		'''Convert Selection To Edges
		'''
		pm.mel.PolySelectConvert(2)


	def b018(self):
		'''Convert Selection To Faces
		'''
		pm.mel.PolySelectConvert(1)


	def b019(self):
		'''Convert Selection To Edge Ring
		'''
		pm.mel.SelectEdgeRingSp()


	@staticmethod
	def setSelectionStyle(ctx):
		'''Set the selection style context.

		Parameters:
			ctx (str): Selection style context. valid: 'marquee', 'lasso', 'paint'.
		'''
		ctx = ctx+'Context'
		if pm.contextInfo(ctx, exists=True):
			pm.deleteUI(ctx)

		if ctx=='marqueeContext':
			ctx = pm.selectContext(ctx)
		elif ctx=='lassoContext':
			ctx = pm.lassoContext(ctx)
		elif ctx=='paintContext':
			ctx = pm.artSelectCtx(ctx)

		pm.setToolTo(ctx)


	def generateUniqueSetName(self, obj=None):
		'''Generate a generic name based on the object's name.

		Parameters:
			obj (str)(obj)(list): The maya scene object to derive a unique name from.

		<objectName>_Set<int>
		'''
		if obj is None:
			obj = pm.ls(sl=1)
		num = ptk.cycle(list(range(99)), 'selectionSetNum')
		name = '{0}_Set{1}'.format(pm.ls(obj, objectsOnly=1, flatten=1)[0].name, num) #ie. pCube1_Set0

		return name


	def creatNewSelectionSet(self, name=None):
		'''Selection Sets: Create a new selection set.
		'''
		if pm.objExists(name):
			self.sb.messageBox('Set with name <hl>{}</hl> already exists.'.format(name))
			return

		else: #create set
			if not name: #name=='set#Set': #generate a generic name based on obj.name
				name = self.generateUniqueSetName()

			pm.sets(name=name, text="gCharacterSet")


	def modifySet(self, name):
		'''Selection Sets: Modify Current by renaming or changing the set members.
		'''
		newName = self.sb.selection.cmb001.currentText()
		if not newName:
			newName = self.generateUniqueSetName()
		name = pm.rename(name, newName)

		if pm.objExists(name):
			pm.sets(name, clear=1)
			pm.sets(name, add=1) #if set exists; clear set and add current selection









#module name
print (__name__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# !/usr/bin/python
# coding=utf-8
from PySide2 import QtCore, QtGui, QtWidgets

from uitk.widgets.attributes import Attributes


class ListWidget(QtWidgets.QListWidget, Attributes):
	'''
	'''

	def __init__(self, parent=None, **kwargs):
		'''
		'''
		super().__init__(parent)

		self.setStyleSheet(parent.styleSheet()) if parent else None

		# self.setViewMode(QtWidgets.QListView.IconMode)
		self.clear()
		self.setAttributes(**kwargs)


	def getItems(self):
		'''
		'''
		return [self.item(i) for i in range(self.count())]


	def getItemsByText(self, text):
		'''
		'''
		return [i for i in self.getItems() if i.text()==text]


	def getItemWidgets(self):
		'''
		'''
		return [self.itemWidget(self.item(i)) for i in range(self.count())]


	def getItemWidgetsByText(self, text):
		'''
		'''
		return [i for i in self.getItemWidgets() if hasattr(i, 'text') and i.text()==text]


	def addItem(self, i):
		'''
		'''
		self.add(i)


	def add(self, w, **kwargs):
		'''Add items to the menu.

		Parameters:
			w () = 
			kwargs:
				show (bool): show the menu.
				insertSeparator (QAction) = insert separator in front of the given action.
		Return:
			the added item object.

		ex.call: menu().add(w='QAction', setText='', insertSeparator=True)
		'''
		try: #get the widget from string class name.
			w = getattr(QtWidgets, w)(self) #ex. QtWidgets.QAction(self) object from string.
		except: #if w is a widget object instead of string.
			if callable(w):
				w = w() #ex. QtWidgets.QAction(self) object.

		typ = w.__class__.__name__

		# if typ=='QListWidgetItem':
			# super().addItem(w)

		if w is not self:
			if typ=='str': #if 'w' is still a string; create a label and use the str value as the label's text.
				lbl = QtWidgets.QLabel(self)
				lbl.setText(w)
				w = lbl
			wItem = QtWidgets.QListWidgetItem(self)
			self.setItemWidget(wItem, w)
			super().addItem(wItem)

		w.__class__.list = property( #add an expandable list to the widget.
			lambda w: w._list if hasattr(w, '_list') else self._addList(w)
		)

		self.setAttributes(w, **kwargs) #set any additional given keyword args for the widget.
		self.resize(self.sizeHint().width(), self.minimumSizeHint().height())
		self.raise_()

		return w


	def _addList(self, w):
		'''Add an expanding list to the given widget.
		Parameters:
			w (obj): 
		'''
		lw = ListWidget(self.parent(), setVisible=False, setObjectName='list')
		w._list = lw
		w._list.prevlist = self #w.parent().parent()
		w.leaveEvent = lambda event, w=w: w._list.leaveEvent(event)
		lw.leaveEvent = lambda event, lw=lw: (self._hideLists(lw)) if not lw.rect().contains(lw.mapFromGlobal(QtGui.QCursor.pos())) else None
		lw.show = lambda lw=lw: lw.__class__.show(lw) if lw.getItemWidgets() else None
		w.installEventFilter(self)
		return w._list


	def _hideLists(self, lw):
		'''Hide the given list and all previous lists in it's hierarchy.
		'''
		while hasattr(lw, 'prevlist'):
			lw.hide()
			lw = lw.prevlist


	def convert(self, items, to='QLabel', **kwargs):
		'''
		Example: self.convert(self.getItems(), 'QPushButton') #construct the list using the existing contents.
		'''
		lst = lambda x: list(x) if isinstance(x, (list, tuple, set, dict)) else [x] #assure 'x' is a list.

		for item in lst(items):
			i = self.indexFromItem(item).row() #get the row as an int from the items QModelIndex.
			item = self.takeItem(i)
			self.add(to, setText=item.text(), **kwargs)


	def eventFilter(self, w, event):
		'''
		'''
		if event.type()==QtCore.QEvent.Enter:
			print(event.type(), w)
			try:
				pos = w.parent().parent().parent().mapFromGlobal(w.mapToGlobal(w.rect().topRight()))
				w.list.move(pos)
				w.list.show()

			except AttributeError as error:
				pass

		if event.type()==QtCore.QEvent.MouseButtonRelease:
			if QtWidgets.QApplication.widgetAt(QtGui.QCursor.pos())==w:
			# if w.rect().contains(w.mapFromGlobal(QtGui.QCursor.pos())):
				try:
					self.itemClicked.emit(w)
				except AttributeError as error:
					pass

		if event.type()==QtCore.QEvent.Leave:
			print ('ef_leaveEvent')
			self.releaseMouse()
			QtWidgets.QApplication.instance().activeWindow().grabMouse()

		return super().eventFilter(w, event)


	def showEvent(self, event):
		'''
		'''
		# for i in self.getItems():
		# 	self.add(i)


	def leaveEvent(self, event):
		'''
		'''
		print ('leaveEvent')
		self.releaseMouse()

# -----------------------------------------------------------------------------









# -----------------------------------------------------------------------------

if __name__ == "__main__":
	import sys
	app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv) #return the existing QApplication object, or create a new one if none exists.

	window = QtWidgets.QWidget()
	lw = ListWidget(window)
	w1 = lw.add('QPushButton', setObjectName='b001', setText='Button 1')
	w1.list.add('list A')
	w2 = lw.add('QPushButton', setObjectName='b002', setText='Button 2')
	w3 = w2.list.add('List B')
	w3.list.add('QPushButton', setObjectName='b004', setText='Button 4')
	lw.add('QPushButton', setObjectName='b003', setText='Button 3')

	print (lw.getItems())
	print (lw.getItemWidgets())

	window.resize(765, 255)
	window.show()
	sys.exit(app.exec_())


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

'''
Promoting a widget in designer to use a custom class:
>	In Qt Designer, select all the widgets you want to replace, 
		then right-click them and select 'Promote to...'. 

>	In the dialog:
		Base Class:		Class from which you inherit. ie. QWidget
		Promoted Class:	Name of the class. ie. "MyWidget"
		Header File:	Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>	Then click "Add", "Promote", 
		and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
'''

# deprecated ---------------------





# !/usr/bin/python
# coding=utf-8
from uitk.slots.max import *
from uitk.slots.animation import Animation



class Animation_max(Animation, Slots_max):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		cmb = self.sb.animation.draggable_header.ctxMenu.cmb000
		items = ['Track View: Curve Editor','Track View: Dope Sheet','Track View: New Track View','Motion Mixer','Pose Mixer','MassFx Tools', 'Dynamics Explorer','Reaction Manager','Walkthrough Assistant']
		cmb.addItems_(items, 'Animation Editors')


	def cmb000(self, index=-1):
		'''Editors
		'''
		cmb = self.sb.animation.draggable_header.ctxMenu.cmb000

		if index>0:
			text = cmb.items[index]
			if text=='Track View: Curve Editor':
				maxEval('macros.run "Track View" "LaunchFCurveEditor"')
			elif text=='Track View: Dope Sheet':
				maxEval('macros.run "Track View" "LaunchDopeSheetEditor"')
			elif text=='Track View: New Track View':
				maxEval('actionMan.executeAction 0 "40278"') #Track View: New Track View
			elif text=='Motion Mixer':
				maxEval('macros.run "Mixer" "ToggleMotionMixerDlg"')
			elif text=='Pose Mixer':
				maxEval('macros.run "CAT" "catPoseMixer"')
			elif text=='MassFx Tools':
				maxEval('macros.run "PhysX" "PxShowToolsWindowMS"')
			elif text=='Dynamics Explorer':
				maxEval('macros.run "PhysX" "OpenDynamicsExplorer"')
			elif text=='Reaction Manager':
				maxEval('macros.run "Animation Tools" "OpenReactionManager"')
			elif text=='Walkthrough Assistant':
				maxEval('macros.run "Animation Tools" "walk_assist"')
			cmb.setCurrentIndex(0)


	def b000(self):
		'''Delete Keys on Selected
		'''
		rt.deleteKeys(rt.selection)


	def getSubAnimControllers(self, node, keyable=False, _result=[]):
		'''Returns a list of all the subanim controllers for a given node.

		Parameters:
			node (obj): The node in which to query controllers of.
			keyable (bool): Return only keyable controllers.
			_result (obj): Recursive call.

		Return:
			(list) A List of sub level animation controllers.

		ex. ctrls = getSubAnimControllers(obj)
		'''
		if rt.iskindof(node, rt.subanim):
			_result.append(node)

		for i in range(1, node.numsubs):
			ctrl = rt.getSubAnim(node, i)
			self.getSubAnimControllers(ctrl, _result)

		if keyable:
			_result = self.getKeyableControllers(_result)

		return _result


	def getKeyableControllers(self, controllers):
		'''Filters a given list for controllers that are keyable.
		You can get the initial list of controllers using getSubAnimControllers.

		Parameters:
			controllers (list): A list of controller objects.

		Return:
			(list) A list of keyable controllers.

		ex. ctrls = getSubAnimControllers(obj)
			kctrls = getKeyableControllers(ctrls)
		'''
		result=[]
		for c in controllers:
			if rt.classof(c.controller)==rt.undefinedClass:
				continue
			if not c.controller.keyable:
				continue
			result.append(c)

		return result


	def AssignController(self, currentController, newController):
		'''Attempts at assigning a given controller.

		Parameters:
			currentController (obj): 
			newController (obj): 

		ex. ctrls = getSubAnimControllers(obj)
			kctrls = getKeyableControllers(ctrls)
			newCtrl = rt.noise_float()
			for k in kctrls:
				AssignController(k.controller, newCtrl)
		'''
		mxsexpr = rt.exprForMAXObject(currentController)
		mxstokens = mxsexpr.split('.')
		controller_to_change = rt.getNodeByName(mxstokens[0][1:]).controller

		for controller_name in mxstokens[1:-2:2]:
			controller_to_change = rt.getPropertyController(controller_to_change, controller_name)

		try:
			rt.setPropertyController(controller_to_change, mxstokens[-2], newController)
		except RuntimeError as error:
			print("# Error: Unable to assign new controller: {} #".format(error))


	def setCurrentFrame(self, frame=1, relative=False, update=True):
		'''Set the current frame on the timeslider.

		Parameters:
		frame (int): Desired from number.
		relative (bool): If True; the frame will be moved relative to 
			it's current position using the frame value as a move amount.
		update (bool): Change the current time, but do not update the world. (default=True)

		Example:
			setCurrentFrame(24, relative=True, update=1)
		'''
		currentTime=0
		if relative:
			currentTime = pm.currentTime(query=True)

		pm.currentTime(currentTime+frame, edit=True, update=update)


	@undo
	def invertSelectedKeyframes(self, time=1, relative=True):
		'''Duplicate any selected keyframes and paste them inverted at the given time.

		Parameters:
			time (int): The desired start time for the inverted keys.
			relative (bool): Start time position as relative or absolute.

		Example: invertSelectedKeyframes(time=48, relative=0)
		'''
		allActiveKeyTimes = pm.keyframe(query=True, sl=True, tc=True) #get times from all selected keys.
		if not allActiveKeyTimes:
			error = '# Error: No keys selected. #'
			print (error)
			return error
		range_ = max(allActiveKeyTimes) - min(allActiveKeyTimes)
		time = time - max(allActiveKeyTimes) if not relative else time

		selection = pm.ls(sl=1, transforms=1)
		for obj in selection:

			keys = pm.keyframe(obj, query=True, name=True, sl=True)
			for node in keys:

				activeKeyTimes = pm.keyframe(node, query=True, sl=True, tc=True)
				for t, rt in zip(activeKeyTimes, reversed(activeKeyTimes)):

					pm.copyKey(node, time=t)
					pm.pasteKey(node, time=rt+range_+time)

					inAngle = pm.keyTangent(node, query=True, time=t, inAngle=True)
					pm.keyTangent(node, edit=True, time=rt+range_+time, inAngle=-inAngle[0])








#module name
print (__name__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# !/usr/bin/python
# coding=utf-8
import sys, os.path

import importlib
import inspect


__package__ = 'uitk'
__version__ = '0.5.9'


def greeting(string, outputToConsole=True):
	'''Format a string using preset variables.

	Parameters:
		string (str): The greeting to format as a string with placeholders using the below keywords. 
			ex. 'Good {hr}! You are using {modver} with {pyver}.'
			{hr} - Gives the current time of day (morning, afternoon, evening)
			{pyver} - The python interpreter version.
			{modver} - This modules version.
		outputToConsole = Print the greeting.

	Return:
		(str)

	Example: greeting('Good {hr}! You are using {modver} with {pyver}.')
	'''
	import datetime
	h = datetime.datetime.now().hour
	hr = 'morning' if 5<=h<12 else 'afternoon' if h<18 else 'evening'

	pyver = 'interpreter v{}.{}.{}'.format(sys.version_info[0], sys.version_info[1], sys.version_info[2])

	modver = 'uitk v{}'.format(__version__)

	result = string.format(hr=hr, pyver=pyver, modver=modver)

	if outputToConsole:
		print (result)
	return result

# --------------------------------------------------------------------------------------------









# --------------------------------------------------------------------------------------------

greeting('Good {hr}! You are using {modver} with {pyver}.')

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# --------------------------------------------------------------------------------------------
# deprecated:
# --------------------------------------------------------------------------------------------


# def greeting(hello=False, mod_version=False, py_version=False):
# 	'''
	# print(greeting(hello=1, mod_version=1, py_version=1))
# 	'''
# 	string=''

# 	if hello: #print greeting
# 		import datetime
# 		hour = datetime.datetime.now().hour
# 		greeting = 'morning' if 5<=hour<12 else 'afternoon' if hour<18 else 'evening'
# 		string+='\nGood {}!'.format(greeting)

# 	if py_version: #print python version
# 		string+='\ninterpreter v{}.{}.{}'.format(sys.version_info[0], sys.version_info[1], sys.version_info[2])

# 	if mod_version:
# 		string+='\ntentacle v{}'.format(__version__)

# 	return string


# !/usr/bin/python
# coding=utf-8
from uitk.slots.maya import *
from uitk.slots.display import Display


class Display_maya(Display, Slots_maya):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		cmb = self.sb.display.draggable_header.ctxMenu.cmb000
		items = ['']
		cmb.addItems_(items, '')


	def cmb000(self, index=-1):
		'''Editors
		'''
		cmb = self.sb.display.draggable_header.ctxMenu.cmb000

		if index>0:
			text = cmb.items[index]
			if text=='':
				pass
			cmb.setCurrentIndex(0)


	def b000(self):
		'''Set Wireframe color
		'''
		pm.mel.objectColorPalette()


	def b001(self):
		'''Toggle Visibility
		'''
		pm.mel.ToggleVisibilityAndKeepSelection()


	def b002(self):
		'''Hide Selected
		'''
		selection = pm.ls(selection=1)
		pm.hide(selection) #pm.mel.HideSelectedObjects()


	def b003(self):
		'''Show Selected
		'''
		selection = pm.ls(selection=1)
		pm.showHidden(selection) #pm.mel.ShowSelectedObjects()


	def b004(self):
		'''Show Geometry
		'''
		pm.mel.hideShow(geometry=1, show=1)


	def b005(self):
		'''Xray Selected
		'''
		pm.mel.eval('''
		string $sel[] = `ls -sl -dag -s`;
		for ($object in $sel) 
			{
			int $xState[] = `displaySurface -query -xRay $object`;
			displaySurface -xRay ( !$xState[0] ) $object;
			}
		''')


	def b006(self):
		'''Un-Xray All
		'''
		pm.mel.eval('''
		string $scene[] = `ls -visible -flatten -dag -noIntermediate -type surfaceShape`;
		for ($object in $scene)
			{
			int $state[] = `displaySurface -query -xRay $object`;
			if ($state[0] == 1)
				{
				displaySurface -xRay 0 $object;
				}
			}
		''')


	def b007(self):
		'''Xray Other
		'''
		pm.mel.eval('''
		//xray all except currently selected
		{
		string $scene[] = `ls -visible -flatten -dag -noIntermediate -type surfaceShape`;
		string $selection[] = `ls -selection -dagObjects -shapes`;
		for ($object in $scene)
			{
			if (!stringArrayContains ($object, $selection))
				{
				int $state[] = `displaySurface -query -xRay $object`;
				displaySurface -xRay ( !$state[0] ) $object;
				}
			}
		}
		''')


	def b008(self):
		'''Filter Objects
		'''
		pm.mel.bt_filterActionWindow()


	def b009(self):
		'''Toggle Material Override
		'''
		from maya.cmds import getPanel #pymel getPanel is broken in ver: 2022.
		currentPanel = getPanel(withFocus=True)
		state = pm.modelEditor(currentPanel, query=1, useDefaultMaterial=1)
		pm.modelEditor(currentPanel, edit=1, useDefaultMaterial=not state)
		mtk.viewportMessage('Default Material Override: <hl>{}</hl>.'.format(state))


	def b011(self):
		'''Toggle Component Id Display
		'''
		index = ptk.cycle([0,1,2,3,4], 'componentID')

		visible = pm.polyOptions(query=1, displayItemNumbers=1)
		if not visible:
			self.sb.messageBox('Nothing selected.')
			return

		dinArray = [(1,0,0,0),(0,1,0,0),(0,0,1,0),(0,0,0,1)]

		if index==4:
			i=0
			for _ in range(4):
				if visible[i]==True:
					pm.polyOptions(relative=1, displayItemNumbers=dinArray[i], activeObjects=1)
				i+=1

		elif visible[index]!=True and index!=4:
			pm.polyOptions(relative=1, displayItemNumbers=dinArray[index], activeObjects=1)

			i=0
			for _ in range(4):
				if visible[i]==True and i!=index:
					pm.polyOptions(relative=1, displayItemNumbers=dinArray[i], activeObjects=1)
				i+=1

		if index==0:
			mtk.viewportMessage("[1,0,0,0] <hl>vertIDs</hl>.")
		elif index==1:
			mtk.viewportMessage("[0,1,0,0] <hl>edgeIDs</hl>.")
		elif index==2:
			mtk.viewportMessage("[0,0,1,0] <hl>faceIDs</hl>.")
		elif index==3:
			mtk.viewportMessage("[0,0,0,1] <hl>compIDs(UV)</hl>.")
		elif index==4:
			mtk.viewportMessage("component ID <hl>Off</hl>.")


	def b012(self):
		'''Wireframe Non Active (Wireframe All But The Selected Item)
		'''
		current_panel = pm.getPanel (withFocus=1)
		state = pm.modelEditor (current_panel, query=1, activeOnly=1)
		pm.modelEditor (current_panel, edit=1, activeOnly=not state)


	def b021(self):
		'''Template Selected
		'''
		pm.toggle(template=1) #pm.toggle(template=1, query=1)









#module name
print (__name__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------
	



# !/usr/bin/python
# coding=utf-8
import os

from PySide2 import QtWidgets

try:
	import pymel.core as pm
except ImportError as error:
	print (__file__, error)

from uitk.slots import Slots
import pythontk as ptk
import mayatk as mtk


class Slots_maya(Slots):
	'''App specific methods inherited by all other slot classes.
	'''
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)


	# ======================================================================
		'UI'
	# ======================================================================

	@classmethod
	def attr(cls, fn):
		'''A decorator for setAttributeWindow (objAttrWindow).
		'''
		def wrapper(self, *args, **kwargs):
			rtn = fn(self, *args, **kwargs)
			self.setAttributeWindow(rtn)
			return rtn
		return wrapper

	def setAttributeWindow(self, obj, inc=[], exc=[], checkableLabel=False, fn=None, fn_args=[], **attributes):
		'''Launch a popup window containing the given objects attributes.

		Parameters:
			obj (str)(obj)(list): The object to get the attributes of, or it's name. If given as a list, only the first index will be used.
			inc (list): Attributes to include. All other will be omitted. Exclude takes dominance over include. Meaning, if the same attribute is in both lists, it will be excluded.
			exc (list): Attributes to exclude from the returned dictionay. ie. ['Position','Rotation','Scale','renderable','isHidden','isFrozen','selected']
			checkableLabel (bool): Set the attribute labels as checkable.
			fn (method) = Set an alternative method to call on widget signal. ex. setParameterValuesMEL
					The first parameter of fn is always the given object. ex. fn(obj, {'attr':<value>})
			fn_args (str)(list): Any additonal args to pass to the given fn.
			attributes (kwargs) = Explicitly pass in attribute:values pairs. Else, attributes will be pulled from mtk.Node.getAttributesMEL for the given obj.

		Example: self.setAttributeWindow(node, attrs, fn=setParameterValuesMEL, 'transformLimits') #set attributes for the Maya command transformLimits.
		Example: self.setAttributeWindow(transform[0], inc=['translateX','translateY','translateZ','rotateX','rotateY','rotateZ','scaleX','scaleY','scaleZ'], checkableLabel=True)
		'''
		try:
			obj = pm.ls(obj)[0]
		except Exception as error:
			return 'Error: {}.setAttributeWindow: Invalid Object: {}'.format(__name__, obj)

		fn = fn if fn else mtk.Node.setAttributesMEL

		if attributes:
			attributes = ptk.Iter.filterDict(attributes, inc, exc, keys=True)
		else:
			attributes = mtk.Node.getAttributesMEL(obj, inc=inc, exc=exc)

		menu = self.objAttrWindow(obj, checkableLabel=checkableLabel, fn=fn, fn_args=fn_args, **attributes)

		if checkableLabel:
			for c in menu.childWidgets:
				if c.__class__.__name__=='QCheckBox':
					attr = getattr(obj, c.objectName())
					c.stateChanged.connect(lambda state, obj=obj, attr=attr: pm.select(attr, deselect=not state, add=1))
					if attr in pm.ls(sl=1):
						c.setChecked(True)

	# -------------------------------------------------------------------------------------------------------------------









#module name
# print (__name__)
# ======================================================================
# Notes
# ======================================================================





#deprecated -------------------------------------

	# @staticmethod
	# def convertToWidget(name):
	# 	'''
	# 	Parameters:
	# 		name (str): name of a Maya UI element of any type. ex. name = pm.mel.eval('$tmp=$gAttributeEditorForm') (from MEL global variable)

	# 	Return:
	# 		(QWidget) None if the object does not exist.
	# 	'''
	# 	import shiboken2
	# 	for f in ('findControl', 'findLayout', 'findMenuItem'):
	# 		ptr = getattr(omui.MQtUtil, f)(name) #equivalent to: ex. omui.MQtUtil.findControl(name)
	# 		if ptr:
	# 			return shiboken2.wrapInstance(long(ptr), QtWidgets.QWidget)


# !/usr/bin/python
# coding=utf-8
from uitk.slots.max import *
from uitk.slots.polygons import Polygons



class Polygons_max(Polygons, Slots_max):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		cmb = self.sb.polygons.draggable_header.ctxMenu.cmb000
		items = ['Bridge','Extrude']
		cmb.addItems_(items, 'Polygon Editors')


	def cmb000(self, index=-1):
		'''Editors
		'''
		cmb = self.sb.polygons.draggable_header.ctxMenu.cmb000

		if index>0:
			text = cmb.items[index]
			if text=='Bridge':
				maxEval('''
				if Ribbon - Modeling.ValidSOMode() and (subObjectLevel == 2 or subObjectLevel == 3) then
				(
					curmod = Modpanel.getcurrentObject()
					if subObjectLevel == 2 then
					(   
					    curmod.popupDialog #BridgeEdge
					)
					else 
					(
					    curmod.popupDialog #BridgeBorder
					)
				)
				''')
			if text=='Extrude':
				maxEval('''
				If subObjectLevel == undefined then Max Modify Mode
				-- default to Face level:
				if subObjectLevel == 0 then subObjectLevel = 4
				local A = modPanel.getCurrentObject()
				if (Filters.Is_This_EditPolyMod A) then
				(
					local level = A.GetMeshSelLevel()
					if (level == #Vertex) then (A.PopupDialog #ExtrudeVertex)
					else if (level == #Edge) then (A.PopupDialog #ExtrudeEdge)
					else if (level == #Face) then (A.PopupDialog #ExtrudeFace)
				)
				else (A.popupDialog #Extrude)
				''')
			cmb.setCurrentIndex(0)


	def tb000(self, state=None):
		'''Merge Vertices
		'''
		tb = self.sb.polygons.tb000

		tolerance = float(tb.ctxMenu.s002.value())
		selection = rt.selection

		if selection:
			for n, obj in enumerate(selection):
				if not self.sb.polygons.progressBar.step(n, len(selection)): #register progress while checking for cancellation:
					break

				vertSelection = rt.getVertSelection(obj)
				if rt.subObjectLevel==1 and vertSelection>1: #merge selected components.
					obj.weldThreshold = tolerance
					rt.polyop.weldVertsByThreshold(obj, vertSelection)

				else: #if object mode. merge all vertices on the selected object.
					rt.polyop.weldVertsByThreshold(obj, obj.verts)
		else:
			self.sb.messageBox('No object selected.')
			return


	def tb001(self, state=None):
		'''Bridge
		'''
		tb = self.sb.polygons.tb001

		divisions = tb.ctxMenu.s003.value()

		for obj in rt.selection:
			obj.EditablePoly.Bridge() #perform bridge
		rt.redrawViews() #redraw changes in viewport


	def tb002(self, state=None):
		'''Combine
		'''
		tb = self.sb.polygons.tb002

		if tb.ctxMenu.chk000.isChecked():
			pass

		sel = rt.selection
		maxEval('''
		j = 1;
		count = sel.count;

		undo off;

		while sel.count > 1 do
		(
			if classof sel != Editable_Poly then converttopoly sel
			(
				polyop.attach sel sel;
			  deleteItem sel (j+1);

			  j += 1;

			  if (j + 1) > sel.count then 
			  (
			      j = 1
			  )
			)
		)
		''')


	def tb003(self, state=None):
		'''Extrude
		'''
		tb = self.sb.polygons.tb003

		keepFacesTogether = tb.ctxMenu.chk002.isChecked() #keep faces/edges together.

		rt.macros.run('Ribbon - Modeling', 'EPoly_Extrude')
		# for obj in rt.selection:
		# 	self.extrudeObject(obj)


	def tb004(self, state=None):
		'''Bevel (Chamfer)
		'''
		tb = self.sb.polygons.tb004

		width = float(tb.ctxMenu.s000.value())

		rt.macros.run('Ribbon - Modeling', 'EPoly_Chamfer')
		# width = float(self.sb.polygons.s000.value())
		# chamfer = True

		# if chamfer:
		# rt.macros.run('Modifiers', 'ChamferMod')
		# else: #bevel
		# 	maxEval('modPanel.addModToSelection (Bevel ()) ui:on')


	def tb005(self, state=None):
		'''Detach
		'''
		tb = self.sb.polygons.tb005

		#rt.macros.run('Ribbon - Modeling', 'GeometryDetach')
		level = rt.subObjectLevel

		
		for obj in rt.selection:
			if level==1: #vertices
				vertices = rt.getVertSelection(obj)
				for v in vertices:
					obj.EditablePoly.breakVerts(v)

			if level==2: #edges
				pass #add function to detach edge. likely as spline.

			if level==4: #faces
				element=rt.polyop.getElementsUsingFace(obj, 1)
				if rt.queryBox('Detach as Element?', title='Detach'): #detach as element
					rt.polyop.detachFaces(obj, element, delete=False, asNode=False)
				else: #detach as separate object
					rt.polyop.detachFaces(obj, element, delete=True, asNode=True)
	

	def tb006(self, state=None):
		'''Inset Face Region
		'''
		tb = self.sb.polygons.tb006

		offset = float(tb.ctxMenu.s001.value())
		maxEval('''
		Try 
		(
			If subObjectLevel == undefined then Max Modify Mode
			local A = modPanel.getCurrentObject()
			if keyboard.shiftpressed then A.popupDialog #Inset
			else A.toggleCommandMode #InsetFace
		)
		Catch()
		''')


	def tb007(self, state=None):
		'''Divide Facet
		'''
		tb = self.sb.polygons.tb007

		dv=u=v=0
		if tb.ctxMenu.chk008.isChecked(): #Split U
			u=2
		if tb.ctxMenu.chk009.isChecked(): #Split V
			v=2

		mode = 0 #The subdivision mode. 0=quads, 1=triangles
		subdMethod = 1 #subdivision type: 0=exponential(traditional subdivision) 1=linear(number of faces per edge grows linearly)
		if tb.ctxMenu.chk010.isChecked(): #tris
			mode=dv=1
			subdMethod=0
		if all([tb.ctxMenu.chk008.isChecked(), tb.ctxMenu.chk009.isChecked()]): #subdivide once into quads
			dv=1
			subdMethod=0
			u=v=0
		#perform operation
		selectedFaces = rt.getFaceSelection()
		# for face in selectedFaces: #when performing polySubdivideFacet on multiple faces, adjacent subdivided faces will make the next face an n-gon and therefore not able to be subdivided. 
		# 	pm.polySubdivideFacet(face, divisions=0, divisionsU=2, divisionsV=2, mode=0, subdMethod=1)


	def tb008(self, state=None):
		'''Boolean Operation
		'''
		tb = self.sb.polygons.tb008

		objects = list(Slots_max.bitArrayToArray(rt.selection))

		if tb.ctxMenu.chk011.isChecked(): #union
			for obj in objects[:-1]:
				objects[-1] + obj

		if tb.ctxMenu.chk012.isChecked(): #difference
			for obj in objects[:-1]:
				objects[-1] - obj

		if tb.ctxMenu.chk013.isChecked(): #intersection
			for obj in objects[:-1]:
				objects[-1] * obj


	def tb009(self, state=None):
		'''Snap Closest Verts
		'''
		tb = self.sb.polygons.tb009

		tolerance = tb.ctxMenu.s005.value()
		freezetransforms = tb.ctxMenu.chk016.isChecked()

		selection = pm.ls(sl=1, objectsOnly=1, type='transform')
		if len(selection)>1:
			obj1, obj2 = selection
			Slots_max.snapClosestVerts(obj1, obj2, tolerance, freezetransforms)
		else:
			self.sb.messageBox('Operation requires at least two selected objects.')
			return


	def b000(self):
		'''Circularize
		'''
		obj = rt.selection[0]
		vertices = rt.polyop.getVertSelection(obj)
		
		self.circularize(vertices)


	def b001(self):
		'''Fill Holes
		'''
		rt.macros.run('Modifiers', 'Cap_Holes')


	def b002(self):
		'''Separate
		'''
		obj = rt.selection[0]

		self.detachElement(obj)


	def b003(self):
		'''Symmetrize
		'''
		pm.mel.Symmetrize()


	def b004(self):
		'''Slice
		'''
		rt.macros.run('Ribbon - Modeling', 'CutsQuickSlice')


	def b009(self):
		'''Collapse Component
		'''
		#--[mesh] 0=object level,1=vertex level,2=edge level,3=face,4=polygon,5=element,
		#--[poly] 0=object level,1=vertex level,2=edge level,3=border,4=polygon,5=element
		
		level = rt.subObjectLevel

		for obj in rt.selection:
			if level == 1: #--vertex level
				obj.EditablePoly.collapse('Vertex')

			if level == 2: #--edge level
				obj.EditablePoly.collapse('Edge')

			if level == 4: #--face level
				obj.EditablePoly.collapse('Face')

		# rt.macros.run('Ribbon - Modeling', 'GeometryCollapse')


	def b012(self):
		'''Multi-Cut Tool
		'''
		self.setSubObjectLevel(4)
		rt.macros.run('Ribbon - Modeling', 'CutsCut')
		# obj = rt.Filters.GetModOrObj()
		# if obj:
		# 	obj.ToggleCommandMode('CutVertex') #cut vertex tool


	def b021(self):
		'''Connect Border Edges
		'''
		mel.eval("performPolyConnectBorders 0;")


	def b022(self):
		'''Attach
		'''
		rt.macros.run('Ribbon - Modeling', 'AttachMode')


	def b028(self):
		'''Quad Draw
		'''
		mel.eval("dR_quadDrawTool;")


	def b032(self):
		'''Poke
		'''
		mel.eval("PokePolygon;")


	def b034(self):
		'''Wedge
		'''
		mel.eval("WedgePolygon;")


	def b038(self):
		'''Assign Invisible
		'''
		mel.eval("polyHole -assignHole 1;")


	def b043(self):
		'''Target Weld
		'''
		self.setSubObjectLevel(1) #set component mode to vertex
		rt.macros.run('Editable Polygon Object', 'EPoly_TargetWeld')
		

	def b045(self):
		'''Re-Order Vertices
		'''
		# symmetryOn = pm.symmetricModelling(query=True, symmetry=True) #query symmetry state
		
		# if symmetryOn:
		# 	pm.symmetricModelling(symmetry=False)
		# mel.eval("setPolygonDisplaySettings(\"vertIDs\");") #set vertex id on
		# mel.eval("doBakeNonDefHistory( 1, {\"pre\"});") #history must be deleted
		# mel.eval("performPolyReorderVertex;") #start vertex reorder ctx


	def b046(self):
		'''Split
		'''
		level = rt.subObjectLevel

		for obj in rt.selection:
			if level==1:
				vertex = rt.polyop.getVertSelection(obj)
				rt.polyop.breakVerts(obj, vertex)

			if level==2:
				position = 0.5

				edges = rt.getEdgeSelection(obj)
				for edge in Slots_max.bitArrayToArray(edges):
					rt.polyop.divideEdge(obj, edge, position)

			if level==4:
				mel.eval("polyChamferVtx 0 0.25 0;")

		rt.redrawViews


	def b047(self):
		'''Insert Edgeloop
		'''
		self.setSubObjectLevel(0)
		rt.macros.run('PolyTools', 'SwiftLoop')


	def b048(self):
		'''Collapse Edgering
		'''
		mel.eval("bt_polyCollapseEdgeRingTool;")


	def b049(self):
		'''Slide Edge Tool
		'''
		mel.eval("SlideEdgeTool;")


	def b050(self):
		'''Spin Edge
		'''
		mel.eval("bt_polySpinEdgeTool;")


	def b051(self):
		'''Offset Edgeloop
		'''
		mel.eval("performPolyDuplicateEdge 0;")


	def b053(self):
		'''Edit Edge Flow
		'''
		mel.eval("PolyEditEdgeFlow;")


	def circularize(self):
		'''Circularize a set of vertices on a circle or an elipse.

		tm = (matrix3 [-0.99596,0.022911,-0.0868241] [-0.0229109,0.870065,0.492404] [0.086824,0.492404,-0.866025] [-18.3751,-66.1508,30.969])
		c = [-18.3751,-66.1508,30.969]
		s = [-123.81,-63.7254,21.7775]
		u = [0.086824,0.492404,-0.866025]

		pCircle = pointCircle c s u 20
		'''
		maxEval('''
		fn pointCircle center startPoint upVector n = (
			rad = distance center startPoint
			dir = normalize (startPoint - center)
			crossVector = normalize (cross (normalize (startPoint - center)) upVector)
			tm = (matrix3 upVector crossVector dir center)

			p3Array = #()

			for i = 1 to n do (
				preRotateX tm (360.0 / n)
				append p3Array ([0,0,rad] * tm)
			)

			return p3Array
		)
		pointCircle()
		''')


	def detachElement(self, obj):
		'''Detach editable_mesh elements into new objects.

		Parameters:
			obj (obj): A polygon object.

		Return:
			(list) detached objects.
		'''
		elementArray = []

		print(obj[0]) #object
		print(obj[6]) #baseObject class TYPE |string|
		print(obj[7]) #isValidNode

		if (obj[4] == rt.Editable_Poly and obj[7]): #or obj[6] == "Shape" or obj[6] == "Geometry" 

			rename = obj[0].name	
			rename += "_ele"
			#~ maxEval("undo \"DetachToElement\" on")
			while ((rt.polyOp.getNumFaces(obj[0])) > 0):
				elementToDetach = rt.polyOp.getElementsUsingFace(obj[0],[1]) #(1)
				rt.polyOp.detachFaces(obj[0], elementToDetach, delete=True, asNode=True, name=rename)
			rt.delete(obj[0])
			elementArray = rt.execute("$"+rename+"*")

			rt.select(elementArray)

		else:
			return 'Error: Object must be an Editable_Poly.'
		
		return elementArray











#module name
print (__name__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# !/usr/bin/python
# coding=utf-8
from uitk.slots.max import *
from uitk.slots.materials import Materials



class Materials_max(Materials, Slots_max):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.randomMat=None

		dh = self.sb.materials.draggable_header
		dh.ctxMenu.add(self.sb.ComboBox, setObjectName='cmb000', setToolTip='Maya Material Editors')
		dh.ctxMenu.add(self.sb.PushButton, setText='Relink Scene Bitmaps', setObjectName='tb003', setToolTip='Repair broken bitmap file links for any scene materials. If no materials are selected, all scene materials will be used.')
		dh.ctxMenu.add(self.sb.PushButton, setText='Relink Library Bitmaps', setObjectName='tb004', setToolTip='Repair broken bitmap file links for all libraries in a given directory.')

		cmb000 = self.sb.materials.draggable_header.ctxMenu.cmb000
		items = ['Material Editor']
		cmb000.addItems_(items, 'Material Editors')


	def cmb000(self, index=-1):
		'''Editors
		'''
		cmb = self.sb.materials.draggable_header.ctxMenu.cmb000

		if index>0:
			text = cmb.items[index]
			if text=='Material Editor':
				maxEval('max mtledit')
			cmb.setCurrentIndex(0)


	def cmb002(self, index=-1):
		'''Material list

		Parameters:
			index (int): parameter on activated, currentIndexChanged, and highlighted signals.
		'''
		cmb = self.sb.materials.cmb002
		b = self.sb.materials_submenu.b003

		mode = cmb.ctxMenu.cmb001.currentText()
		if mode=='Scene Materials':
			materials = self.getSceneMaterials()

		elif mode=='ID Map Materials':
			materials = self.getSceneMaterials(startingWith=['ID_'])

		if mode=='Favorite Materials':
			currentMats = {m:m for m in ['standardMaterial']}
		else:
			currentMats = {mat.name:mat for mat in sorted(list(set(materials))) if hasattr(mat, 'name')}

		cmb.addItems_(currentMats, clear=True)

		#create and set icons with color swatch
		for i, mat in enumerate(cmb.items):
			icon = self.getColorSwatchIcon(mat)
			cmb.setItemIcon(i, icon) if icon else None

		#set submenu assign material button attributes
		b.setText('Assign '+cmb.currentText())
		icon = self.getColorSwatchIcon(cmb.currentText(), [15, 15])
		b.setIcon(icon) if icon else None
		b.setMinimumWidth(b.minimumSizeHint().width()+25)
		b.setVisible(True if cmb.currentText() else False)


	def tb000(self, state=None):
		'''Select By Material Id
		'''
		tb = self.sb.materials.tb000

		mat = self.sb.materials.cmb002.currentData()
		if not mat:
			self.sb.messageBox('No Material Selection.')
			return

		shell = tb.ctxMenu.chk005.isChecked() #Select by material: shell
		invert = tb.ctxMenu.chk006.isChecked() #Select by material: invert
		allObjects = tb.ctxMenu.chk003.isChecked() #Search all scene objects

		objects = rt.selection if not allObjects else None
		
		self.selectByMaterialID(mat, objects, shell=shell, invert=invert)


	def tb002(self, state=None):
		'''Assign Material
		'''
		tb = self.sb.materials.tb002

		selection = rt.selection

		assignCurrent = tb.ctxMenu.chk007.isChecked()
		assignRandom = tb.ctxMenu.chk008.isChecked()
		assignNew = tb.ctxMenu.chk009.isChecked()

		if assignRandom: #Assign New random mat ID
			if selection:
				mat = self.createRandomMaterial(prefix='ID_')
				self.assignMaterial(selection, mat)

				#delete previous shader
				if self.randomMat:
					self.randomMat = None #replace with standard material

				self.randomMat = mat

				if self.sb.materials.tb001.menu_.chk001.isChecked(): #ID map materials mode.
					self.cmb002() #refresh the combobox
				else:
					self.sb.materials.tb001.menu_.chk001.setChecked(True) #set comboBox to ID map mode. toggling the checkbox refreshes the combobox.
				self.sb.materials.cmb002.setCurrentItem(mat.name) #set the comboBox index to the new mat #self.cmb002.setCurrentIndex(self.cmb002.findText(name))
			else:
				self.sb.messageBox('No valid object/s selected.')
				return

		elif assignCurrent: #Assign current mat
			if isinstance(mat, str): #new mat type as a string:
				self.assignMaterial(selection, rt.Standard(name=mat))
			else: #existing mat object:
				self.assignMaterial(selection, mat)

		elif assignNew: #Assign New Material
			pass

		rt.redrawViews()


	def tb003(self, state=None):
		'''Relink Scene Bitmaps
		'''
		tb = self.sb.materials.draggable_header.ctxMenu.tb003
		if state=='setMenu':
			tb.ctxMenu.add('QLineEdit', setPlaceholderText='Set Bitmaps Directory:', setText=r'\\m3trik-Server\NAS\Graphics\_materials', setObjectName='l000', setToolTip='Location to search for missing bitmaps.') #
			return

		mat_dir = tb.ctxMenu.l000.text()
		mats = self.getNodesSME(selected=True) #find bitmaps for any currently selected nodes in the slate material editor, else relink all scene nodes.
		if not mats:
			mats = None

		self.relinkSceneBitmaps(mat_dir, mats=mats, replaceTxWithTif=True)


	def tb004(self, state=None):
		'''Relink Material Library Bitmaps
		'''
		tb = self.sb.materials.draggable_header.ctxMenu.tb004
		if state=='setMenu':
			tb.ctxMenu.add('QLineEdit', setPlaceholderText='Set Bitmaps Directory:', setText=r'\\m3trik-Server\NAS\Graphics\_materials', setObjectName='l001', setToolTip='Location to search for missing bitmaps.') #
			tb.ctxMenu.add('QLineEdit', setPlaceholderText='Set Material Library Directory:', setText=r'\\m3trik-Server\NAS\Graphics\_materials\libraries', setObjectName='l002', setToolTip='Location of material libraries.') #
			return

		library_dir = tb.ctxMenu.l001.text()
		mat_dir = tb.ctxMenu.l002.text()

		self.relinkMatLibBitmaps(library_dir, mat_dir, replaceTxWithTif=True)


	def lbl000(self):
		'''Open material in editor
		'''
		try:
			mat = self.sb.materials.cmb002.currentData() #get the mat obj from cmb002
			rt.select(mat)
		except Exception as error:
			self.sb.messageBox('No stored material or no valid object selected.')
			return

		#open the slate material editor
		if not rt.SME.isOpen():
			rt.SME.open()

		#create a temp view in the material editor
		if rt.SME.GetViewByName('temp'):
			rt.SME.DeleteView(rt.SME.GetViewByName('temp'), False)
		index = rt.SME.CreateView('temp')
		view = rt.SME.GetView(index)

		#show node and corresponding parameter rollout
		node = view.CreateNode(mat, rt.point2(0, 0))
		rt.SME.SetMtlInParamEditor(mat)

		rt.redrawViews()


	def lbl001(self, setEditable=True):
		'''Rename Material: Set cmb002 as editable and disable wgts.
		'''
		cmb = self.sb.materials.cmb002

		if setEditable:
			self._mat = self.sb.materials.cmb002.currentData()
			cmb.setEditable(True)
			self.sb.toggleWidgets(self.sb.materials, setDisabled='b002,lbl000,tb000,tb002')
		else:
			mat = self._mat
			newMatName = cmb.currentText()
			self.renameMaterial(mat, newMatName)
			cmb.setEditable(False)
			self.sb.toggleWidgets(self.sb.materials, setEnabled='b002,lbl000,tb000,tb002')


	def lbl002(self):
		'''Delete Material
		'''
		cmb = self.sb.materials.cmb002

		mat = self.sb.materials.cmb002.currentData()
		default_mat = rt.Standard(mat, name="Default Material") #replace with standard material

		index = cmb.currentIndex()
		cmb.setItemText(index, default_mat.name) #self.sb.materials.cmb002.removeItem(index)


	def lbl003(self):
		'''Delete Unused Materials
		'''
		defaultMaterial = rt.Standard(name='Default Material')
		
		for mat in rt.sceneMaterials:
			nodes = rt.refs().dependentnodes(mat) 
			if nodes.count==0:
				rt.replaceinstances(mat, defaultMaterial)
				
			rt.gc()
			rt.freeSceneBitmaps()


	def b002(self):
		'''Set Material: Set the Currently Selected Material as the currentMaterial.
		'''
		try: 
			obj = rt.selection[0]
		except IndexError:
			self.sb.messageBox('Nothing selected.')
			return

		mat = self.getMaterial()

		self.sb.materials.cmb001.setCurrentIndex(0) #set the combobox to show all scene materials
		self.cmb002() #refresh the materials list comboBox
		self.sb.materials.cmb002.setCurrentIndex(self.sb.materials.cmb002.items.index(mat.name()))


	def getColorSwatchIcon(self, mat, size=[20, 20]):
		'''Get an icon with a color fill matching the given materials RBG value.

		Parameters:
			mat (obj)(str): The material or the material's name.
			size (list): Desired icon size. [width, height]

		Return:
			(obj) pixmap icon.
		'''
		try:
			mat = next(m for m in self.getSceneMaterials() if m.name==mat) if isinstance(mat, (str)) else mat #get the mat object if a string name is given.
			r = int(mat.diffuse.r) #convert from float value
			g = int(mat.diffuse.g)
			b = int(mat.diffuse.b)
			pixmap = QtGui.QPixmap(size[0],size[1])
			pixmap.fill(QtGui.QColor.fromRgb(r, g, b))

			return QtGui.QIcon(pixmap)

		except (StopIteration, AttributeError):
			pass


	def renameMaterial(self, mat, newMatName):
		'''Rename Material
		'''
		cmb = self.sb.materials.cmb002 #scene materials

		curMatName = mat.name
		if curMatName!=newMatName:
			cmb.setItemText(cmb.currentIndex(), newMatName)
			try:
				curMatName = newMatName
			except RuntimeError as error:
				cmb.setItemText(cmb.currentIndex(), str(error.strip('\n')))


	def selectByMaterialID(self, material=None, objects=None, shell=False, invert=False):
		'''Select By Material Id
	
		material (obj): The material to search and select for.
		objects (list): Faces or mesh objects as a list. If no objects are given, all geometry in the scene will be searched.
		shell (bool): Select the entire shell.
		invert (bool): Invert the final selection.

		#ex call:
		selectByMaterialID(material)
		'''
		if rt.getNumSubMtls(material): #if not a multimaterial
			self.sb.messageBox('No valid stored material. If material is a multimaterial, select a submaterial.')
			return

		if not material:
			material = self.getMaterial()

		if not objects: #if not selection; use all scene geometry
			objects = rt.geometry

		for obj in objects:
			if not any([rt.isKindOf(obj, rt.Editable_Poly), rt.isKindOf(obj, rt.Editable_mesh)]):
				print('Error: '+str(obj.name)+' skipped. Operation requires an Editable_Poly or Editable_mesh.')
			else:
				if shell: #set to base object level
					rt.modPanel.setCurrentObject(obj.baseObject)
				else: #set object level to face
					Slots_max.setSubObjectLevel(4)
				m = obj.material
				multimaterial = rt.getNumSubMtls(m)

				same=[] #list of faces with the same material
				other=[] #list of all other faces

				faces = list(range(1, obj.faces.count))
				for f in faces:
					if multimaterial:
						try: #get material from face
							index = rt.GetFaceId_(obj, f) #Returns the material ID of the specified face.
						except RuntimeError: #try procedure for polygon object
							index = rt.polyop.GetFaceId_(obj, f) #Returns the material ID of the specified face.
						m = obj.material[index-1] #m = rt.getSubMtl(m, id) #get the material using the ID_ index (account for maxscript arrays starting at index 1)

					if m==material: #single material
						if shell: #append obj to same and break loop
							same.append(obj)
							break
						else: #append face ID to same
							same.append(f)
					else:
						if shell: #append obj to other and break loop
							other.append(obj)
							break
						else: #append face ID to other
							other.append(f)

				if shell:
					if invert:
						(rt.select(i) for i in other)
					else:
						(rt.select(i) for i in same)
				else:
					if invert:
						try:
							rt.setFaceSelection(obj, other) #select inverse of the faces for editable mesh.
						except RuntimeError:
							rt.polyop.setFaceSelection(obj, other) #select inverse of the faces for polygon object.
					else:
						try:
							rt.setFaceSelection(obj, same) #select the faces for editable mesh.
						except RuntimeError:
							rt.polyop.setFaceSelection(obj, same) #select the faces for polygon object.
				# print same
				# print other


	def getSceneMaterials(self.startingWith=['']):
		'''Get All Materials from the current scene.

		Parameters:
			startingWith (list): Filters material names starting with any of the strings in the given list. ie. ['ID_']
		Return:
			(list) materials.
		'''
		materials=[] #get any scene material that does not start with 'Material'
		for mat in rt.sceneMaterials:
			if rt.getNumSubMtls(mat): #if material is a submaterial; search submaterials
				for i in range(1, rt.getNumSubMtls(mat)+1):
					subMat = rt.getSubMtl(mat, i)
					if subMat and filter(subMat.name.startswith, startingWith):
						materials.append(subMat)
			elif filter(mat.name.startswith, startingWith):
				materials.append(mat)

		return materials


	def getMaterial(self, obj=None, face=None):
		'''Get the material from the given object or face components.

		Parameters:
			obj (obj): Mesh object.
			face (int): Face number.
		Return:
			(obj) material
		'''
		if not obj:
			selection = rt.selection
			if not selection:
				self.sb.messageBox('Nothing selected. Select an object face, or choose the option: current material.')
				return
			obj = selection[0]

		mat = obj.material #get material from selection

		if rt.subObjectLevel==4: #if face selection check for multimaterial
			if rt.getNumSubMtls(mat): #if multimaterial; use selected face to get material ID
				if face is None:
					face = self.bitArrayToArray(rt.getFaceSelection(obj))[0] #get selected face

				if rt.classOf(obj)==rt.Editable_Poly:
					ID_ = rt.polyop.GetFaceId_(obj, face) #Returns the material ID of the specified face.
				else:
					try:
						ID_ = rt.GetFaceId_(obj, face) #Returns the material ID of the specified face.
					except RuntimeError:
						self.sb.messageBox('Object must be of type Editable_Poly or Editable_mesh.')
						return

				mat = rt.getSubMtl(mat, ID_) #get material from mat ID

		return mat


	def createRandomMaterial(self, name='', prefix=''):
		'''Creates a random material.

		Parameters:
			name (str): material name.
			prefix (str): Optional string to be appended to the beginning of the name.

		Return:
			(obj) material
		'''
		import random
		rgb = [random.randint(0, 255) for _ in range(3)] #generate a list containing 3 values between 0-255

		name = '{}{}_{}_{}_{}'.format(prefix, name, str(rgb[0]), str(rgb[1]), str(rgb[2]))

		#create shader
		mat = rt.StandardMaterial()
		mat.name = name
		mat.diffuse = rt.color(rgb[0], rgb[1], rgb[2])

		return mat


	def assignMaterial(self, objects, mat):
		'''Assign Material

		objects (list): Faces or mesh objects as a list.
		material (obj): The material to search and select for.
		'''
		if not mat:
			self.sb.messageBox('Material Not Assigned. No material given.')
			return

		for obj in objects:
			if rt.getNumSubMtls(mat): #if multimaterial
				mat.materialList.count = mat.numsubs+1 #add slot to multimaterial
				mat.materialList[-1] = material #assign new material to slot
			else:
				obj.material = mat

		rt.redrawViews()


	def getMaterialBitmaps(self, mats=None, missing=False, processChildren=True):
		'''Get any bitmaps from a given material(s), or from all scene materials.

		Parameters:
			mats (obj)(list): Mat object or list of mat objects. If None is given, all bitmap textures in the scene are used.
			missing (bool): Return only filenames from missing bitmaps.
			processChildren (bool): Child scene nodes are also searched as part of the Animatable or Reference hierarchy.

		Return:
			(list) material bitmaps.
		'''
		if mats is None:
			result = list(rt.getClassInstances(rt.BitmapTexture, processChildren=False))
		else:
			result=[]
			for mat in list(mats):
				result+=list(rt.getClassInstances(rt.BitmapTexture, target=mat, processChildren=processChildren))

		if missing: #get only the names from bitmaps that are missing.
			result = {bitmap:bitmap.filename.rsplit('\\')[-1]
			for bitmap in bitmaps
				if not rt.doesFileExist(bitmap.filename)}

		return result


	def getBitmapFilenames(self, bitmaps=None, missing=False, returnType=list):
		'''Get the file paths for the given bitmaps. If no bitmaps are given all bitmaps in the scene will be used.

		Parameters:
			bitmaps (list): A list of bitmaps. If no bitmaps are given all bitmaps in the scene will be used.
			missing (bool): Return only filenames from missing bitmaps.
			returnType (type) = Valid (list (default), dict).

		Return:
			dependant on returnType flag.
			(dict) {bitmap object:filepath}
			(list) [filepath]
		'''
		import fnmatch

		if not bitmaps: #if no bitmaps are given, use all scene bitmaps.
			bitmaps = self.getMaterialBitmaps(missing=missing)

		result = {bitmap:bitmap.filename.rsplit('\\')[-1] for bitmap in bitmaps}

		if returnType is list:
			result = result.values()

		return result


	def setBitmapFilenames(self, dict_, reload=False):
		'''Set the file paths for the given bitmaps. Bitmaps are given as a dict of bitmaps as keys, and filenames as values.

		Parameters:
			dict_ (dict): A dict of bitmaps as keys, and filenames as values.
			reload (bool): Refresh the bitmap node after updating the path.
		'''
		for bitmap,name in dict_.items():
			bitmap.filename = name
			if reload:
				bitmap.reload()


	def relinkBitmaps(self, dir_, bitmaps=None, replaceTxWithTif=False):
		'''Find the first valid path in the given dir for each bitmap in a given dict. If no bitmaps are given all bitmaps in the scene will be used.

		Parameters:
			dir_ (str): The parent dir to recursively search for files in.
			bitmaps (dict): A dict of bitmaps as keys, and filenames as values. If no bitmaps are given all bitmaps in the scene will be used.
			replaceTxWithTif (bool): Look instead for a .tif file of the same name, to replace a previoud .tx format.

		Return:
			(dict) Any bitmaps that are not found. {bitmap object:filename}
		'''
		import fnmatch, os

		bitmaps = self.getBitmapFilenames(bitmaps, missing=True, returnType=dict)

		if replaceTxWithTif:
			for bitmap,filename in bitmaps.items():
				if filename.endswith('.tx'):
					bitmaps[bitmap] = filename.replace('.tx', '.tif')

		print('Relinking bitmaps in {} ..'.format(dir_))
		result={}
		for root, dirnames, filenames in os.walk(dir_):

			for bitmap, pattern in bitmaps.items():
				for filename in fnmatch.filter(filenames, pattern):
					path = os.path.join(root, filename)
					result[bitmap] = path
					bitmaps.pop(bitmap, None)
					print ('# Result: {}: {}: {} #'.format(filename, bitmap.name, path))

		self.setBitmapFilenames(result, reload=True)

		if bitmaps:
			for bitmap,filename in bitmaps.items():
				print ('# Error: {}: {} #'.format(bitmap.name, filename))


	def relinkMatLibBitmaps(self, library_dir, mat_dir, replaceTxWithTif=False):
		'''Repair broken bitmap file links for all libraries in a given directory.

		Parameters:
			library_dir (str): A path to a directory containing the library files.
			mat_dir (str): A path to a directory containing the material dependancies.
		'''
		import os

		bitmaps=[]; tempLibs={}
		for root, dirnames, filenames in os.walk(library_dir):
			for filename in filenames:
				if filename.endswith('.mat'):
					try:
						path = os.path.join(root, filename)
						tempLib = rt.loadTempMaterialLibrary(path)

						for bitmap in self.getMaterialBitmaps(tempLib):
							bitmaps.append(bitmap)

						tempLibs[tempLib] = path
					except Exception as e:
						print ('# Error: {}: {} #'.format(filename, e))

		if bitmaps:
			self.relinkBitmaps(mat_dir, bitmaps, replaceTxWithTif=replaceTxWithTif)
			{rt.saveTempMaterialLibrary(lib, name) for lib,name in tempLibs.items()}


	def relinkSceneBitmaps(self, mat_dir, mats=None, replaceTxWithTif=False):
		'''Repair broken bitmap file links.  If no mats are given, all scene bitmaps will be used.

		Parameters:
			mats (obj)(list): Specify material(s) to get bitmaps for. If none are given, all scene materials will be used.
			replaceTxWithTif (bool): Look instead for a .tif file of the same name, to replace a previoud .tx format.
		'''
		bitmaps = self.getMaterialBitmaps(mats)

		if bitmaps:
			self.relinkBitmaps(mat_dir, bitmaps, replaceTxWithTif=True)


	def getNodesSME(self, nodeType=None, selected=False):
		'''Get any nodes in the slate material editor that are currently selected.

		Parameters:
			nodeType (obj)(str): The type of node to filter the results for. If 'query' is given, the node type will be returned. ie. rt.VRayMtl()
			selected (bool): When True, only currently selected nodes will be returned.
		Return:
			(list) nodes or node types.
		'''
		view = rt.SME.getView(rt.SME.activeView)

		result=[]
		for i in range(1, view.GetNumNodes()):
			node = view.getNode(i)
			if all((nodeType, rt.isKindOf(node, nodeType), selected, node.selected)): #get selected of type.
				result.append(node)

			elif all((nodeType, nodeType=='query', selected, node.selected)): #query type of selected.
				result.append(rt.classOf(node))

			elif selected and node.selected: #get selected.
				result.append(node)

			elif all((nodeType, nodeType!='query', rt.isKindOf(node, nodeType))): #get node of type.
				result.append(node)

			elif nodeType and nodeType=='query': #query type.
				result.append(rt.classOf(node))

			else: #get all SME nodes.
				result.append(node)

		return result


	def compareMats(self, obj1, obj2):
		'''Compare material names. unlike other properties, this is a simple true/false comparison, it doesn't find 'similar' names.

		Parameters:
			obj1 (obj): 
			obj2 (obj): 
		'''
		maxEval('''
			m1 = obj1.material
			m2 = obj2.material
			
			if m1 != undefined and m2 != undefined then --verify both objects have a material assigned
			(
				if m1.name == m2.name then 
				(
					dbgSelSim ("  Material match on object: '" + obj1.name + "' with '" + obj2.name + "'")
					true	--check if material names are the same, if they are, return a true value
				)
				else false --uh oh. material names aren't the same. return false.
			)
			else false	--one or both objects do not have a material assigned. returning false.
			''')









#module name
print (__name__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# deprecated: -----------------------------------


	# @property
	# def currentMat(self):
	# 	'''Get the current material using the current index of the materials combobox.
	# 	'''
	# 	text = self.sb.materials.cmb002.currentText()

	# 	try:
	# 		result = self.currentMats[text]
	# 	except:
	# 		result = None

	# 	return result


	# def chk007(self, state=None):
	# 	'''Assign Material: Current
	# 	'''
	# 	self.sb.materials.tb002.setText('Assign Current')


	# def chk008(self, state=None):
	# 	'''Assign Material: Random
	# 	'''
	# 	self.sb.materials.tb002.setText('Assign Random')


	# def chk009(self, state=None):
	# 	'''Assign Material: New
	# 	'''
	# 	self.sb.materials.tb002.setText('Assign New')


# !/usr/bin/python
# coding=utf-8
from uitk.slots import Slots



class Symmetry(Slots):
	'''
	'''
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		'''
		'''
		dh = self.sb.symmetry.draggable_header
		dh.ctxMenu.add(self.sb.ComboBox, setObjectName='cmb000', setToolTip='')


	def draggable_header(self, state=None):
		'''Context menu
		'''
		dh = self.sb.symmetry.draggable_header


	def chk000(self, state=None):
		'''Symmetry X
		'''
		self.sb.toggleWidgets(setUnChecked='chk001,chk002')
		self.setSymmetry(state, 'x')


	def chk001(self, state=None):
		'''Symmetry Y
		'''
		self.sb.toggleWidgets(setUnChecked='chk000,chk002')
		self.setSymmetry(state, 'y')


	def chk002(self, state=None):
		'''Symmetry Z
		'''
		self.sb.toggleWidgets(setUnChecked='chk000,chk001')
		self.setSymmetry(state, 'z')


	def chk004(self, state=None):
		'''Symmetry: Object
		'''
		self.sb.symmetry.chk005.setChecked(False) #uncheck symmetry:topological









# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



#deprecated:
		#sync widgets
		# self.sb.syncWidgets(self.sb.transform.menu_.chk000, self.sb.transform_submenu.chk000, attributes='setChecked')
		# self.sb.syncWidgets(self.sb.transform.menu_.chk001, self.sb.transform_submenu.chk001, attributes='setChecked')
		# self.sb.syncWidgets(self.sb.transform.menu_.chk002, self.sb.transform_submenu.chk002, attributes='setChecked')


# !/usr/bin/python
# coding=utf-8
from uitk.slots import Slots



class Normals(Slots):
	'''
	'''
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		'''
		'''
		dh = self.sb.normals.draggable_header
		dh.ctxMenu.add(self.sb.ComboBox, setObjectName='cmb000', setToolTip='')

		tb000 = self.sb.normals.tb000
		tb000.ctxMenu.add('QSpinBox', setPrefix='Display Size: ', setObjectName='s001', setMinMax_='1-100 step1', setValue=1, setToolTip='Normal display size.')

		tb001 = self.sb.normals.tb001
		tb001.ctxMenu.add('QSpinBox', setPrefix='Angle: ', setObjectName='s002', setMinMax_='0-180 step1', setValue=0, setToolTip='The normal angle in degrees.')
		tb001.ctxMenu.add('QCheckBox', setText='Harden Creased Edges', setObjectName='chk005', setChecked=True, setToolTip='Soften all non-creased edges.')
		tb001.ctxMenu.add('QCheckBox', setText='Harden UV Borders', setObjectName='chk006', setChecked=True, setToolTip='Harden UV shell border edges.')
		tb001.ctxMenu.add('QCheckBox', setText='Soften All Other', setObjectName='chk004', setChecked=True, setToolTip='Soften all non-hard edges.')
		tb001.ctxMenu.add('QCheckBox', setText='Soft Edge Display', setObjectName='chk007', setChecked=True, setToolTip='Turn on soft edge display for the object.')

		tb002 = self.sb.normals.tb002
		tb002.ctxMenu.add('QSpinBox', setPrefix='Angle: ', setObjectName='s000', setMinMax_='0-180 step1', setValue=60, setToolTip='Angle degree.')

		tb003 = self.sb.normals.tb003
		tb003.ctxMenu.add('QCheckBox', setText='Lock', setObjectName='chk002', setChecked=True, setToolTip='Toggle Lock/Unlock.')
		tb003.ctxMenu.add('QCheckBox', setText='All', setObjectName='chk001', setChecked=True, setToolTip='Lock/Unlock All.')
		tb003.ctxMenu.chk002.toggled.connect(lambda state, w=tb003.ctxMenu.chk002: w.setText('Lock') if state else w.setText('Unlock')) 

		tb004 = self.sb.normals.tb004
		tb004.ctxMenu.add('QCheckBox', setText='By UV Shell', setObjectName='chk003', setChecked=True, setToolTip='Average the normals of each object\'s faces per UV shell.')


	def draggable_header(self, state=None):
		'''Context menu
		'''
		dh = self.sb.normals.draggable_header



# !/usr/bin/python
# coding=utf-8
from uitk.slots import Slots



class Duplicate(Slots):
	'''
	'''
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		'''
		'''
		ctx = self.sb.duplicate.draggable_header.ctxMenu
		if not ctx.containsMenuItems:
			ctx.add(self.sb.ComboBox, setObjectName='cmb000', setToolTip='')


	def draggable_header(self, state=None):
		'''Context menu
		'''
		dh = self.sb.duplicate.draggable_header


	def radialArray(self):
		'''Radial Array: Reset
		'''
		self.chk015() #calling chk015 directly from valueChanged would pass the returned spinbox value to the create arg


	def duplicateArray(self):
		'''Duplicate: Reset
		'''
		self.chk016() #calling chk015 directly from valueChanged would pass the returned spinbox value to the create arg


	def chk007(self, state=None):
		'''Duplicate: Translate To Components
		'''
		if self.sb.duplicate_linear.chk007.isChecked():
			self.sb.toggleWidgets(setEnabled='chk008,b034,cmb001', setDisabled='chk000,chk009,s005')
			self.b008()
		else:
			self.sb.toggleWidgets(setDisabled='chk008,b034,cmb001', setEnabled='chk000,chk009,s005')


	def chk011(self, state=None):
		'''Radial Array: Instance/Duplicate Toggle
		'''
		self.chk015() #calling chk015 directly from valueChanged would pass the returned spinbox value to the create arg


	def chk012(self, state=None):
		'''Radial Array: X Axis
		'''
		self.sb.toggleWidgets(setChecked='chk012', setUnChecked='chk013,chk014')
		self.chk015()


	def chk013(self, state=None):
		'''Radial Array: Y Axis
		'''
		self.sb.toggleWidgets(setChecked='chk013', setUnChecked='chk012,chk014')
		self.chk015()


	def chk014(self, state=None):
		'''Radial Array: Z Axis
		'''
		self.sb.toggleWidgets(setChecked='chk014', setUnChecked='chk012,chk013')
		self.chk015()


	def b002(self):
		'''Duplicate: Create
		'''
		self.sb.duplicate_linear.chk016.setChecked(False) #must be in the false unchecked state to catch the create flag in chk015
		self.chk016(create=True)


	def b003(self):
		'''Radial Array: Create
		'''
		self.sb.duplicate_radial.chk015.setChecked(False) #must be in the false unchecked state to catch the create flag in chk015
		self.chk015(create=True)


	def b006(self):
		'''
		'''
		self.sb.parent().setUi('duplicate_linear')
		self.sb.duplicate_linear.s002.valueChanged.connect(self.duplicateArray) #update duplicate array
		self.sb.duplicate_linear.s003.valueChanged.connect(self.duplicateArray)
		self.sb.duplicate_linear.s004.valueChanged.connect(self.duplicateArray)
		self.sb.duplicate_linear.s005.valueChanged.connect(self.duplicateArray)
		self.sb.duplicate_linear.s007.valueChanged.connect(self.duplicateArray) 
		self.sb.duplicate_linear.s008.valueChanged.connect(self.duplicateArray)
		self.sb.duplicate_linear.s009.valueChanged.connect(self.duplicateArray)
		self.sb.duplicate_linear.s010.valueChanged.connect(self.duplicateArray) 
		self.sb.duplicate_linear.s011.valueChanged.connect(self.duplicateArray)
		self.sb.duplicate_linear.s012.valueChanged.connect(self.duplicateArray)


	def b007(self):
		'''
		'''
		self.sb.parent().setUi('duplicate_radial')
		self.sb.duplicate_radial.s000.valueChanged.connect(self.radialArray) #update radial array
		self.sb.duplicate_radial.s001.valueChanged.connect(self.radialArray) 


	def b008(self):
		'''Add Selected Components To cmb001
		'''
		cmb = self.sb.duplicate_linear.cmb001

		selection = pm.ls(selection=1, flatten=1)

		for obj in selection:
			cmb.add(obj)



# !/usr/bin/python
# coding=utf-8
import os

from PySide2 import QtGui, QtWidgets, QtCore

try: #Blender dependancies
	import bpy

except ImportError as error:
	print(__file__, error)

from uitk.slots import Slots



class Slots_blender(Slots):
	'''App specific methods inherited by all other slot classes.
	'''
	def __init__(self, *args, **kwargs):
		Slots.__init__(self, *args, **kwargs)


	def undo(fn):
		'''A decorator to place a function into Maya's undo chunk.
		Prevents the undo queue from breaking entirely if an exception is raised within the given function.

		Parameters:
			fn (obj): The decorated python function that will be placed into the undo que as a single entry.
		'''
		def wrapper(*args, **kwargs):
			# pm.undoInfo(openChunk=True)
			rtn = fn(*args, **kwargs)
			# pm.undoInfo(closeChunk=True)
			return rtn
		return wrapper









#module name
# print (__name__)
# ======================================================================
# Notes
# ======================================================================





#deprecated -------------------------------------




# !/usr/bin/python
# coding=utf-8
from PySide2 import QtWidgets, QtCore

from uitk.widgets.attributes import Attributes
from uitk.widgets.text import RichText, TextOverlay
from uitk.widgets.menu import MenuInstance
from uitk.widgets.pushButton_optionBox import PushButton_optionBox


class PushButton(QtWidgets.QPushButton, MenuInstance, Attributes, RichText, TextOverlay):
	'''
	'''
	def __init__(self, parent=None, showMenuOnMouseOver=False, **kwargs):
		QtWidgets.QPushButton.__init__(self, parent)

		self.setStyleSheet(parent.styleSheet()) if parent else None

		self.menu_.position = 'topRight'
		self.showMenuOnMouseOver = showMenuOnMouseOver

		self.setAttribute(QtCore.Qt.WA_SetStyle) #Indicates that the widget has a style of its own.

		#override built-ins
		self.text = self.richText
		self.setText = self.setRichText
		self.sizeHint = self.richTextSizeHint

		self.optionBox = None

		self.setAttributes(**kwargs)


	def enterEvent(self, event):
		'''
		Parameters:
			event = <QEvent>
		'''
		if self.showMenuOnMouseOver:
			self.menu_.show()

		QtWidgets.QPushButton.enterEvent(self, event)


	def mousePressEvent(self, event):
		'''
		Parameters:
			event = <QEvent>
		'''
		if event.button()==QtCore.Qt.RightButton:
			self.ctxMenu.show()

		QtWidgets.QPushButton.mousePressEvent(self, event)


	def leaveEvent(self, event):
		'''
		Parameters:
			event = <QEvent>
		'''
		if self.showMenuOnMouseOver:
			self.menu_.hide()

		QtWidgets.QPushButton.leaveEvent(self, event)


	def createOptionBox(self):
		'''
		'''
		self.optionBox = PushButton_optionBox(self) #create an option box
		self.optionBox.create()


	def showEvent(self, event):
		'''
		Parameters:
			event = <QEvent>
		'''
		if self.ctxMenu.containsMenuItems:
			if not self.optionBox:
				self.createOptionBox()

		QtWidgets.QPushButton.showEvent(self, event)









if __name__ == "__main__":
	import sys
	from PySide2.QtCore import QSize
	app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv) #return the existing QApplication object, or create a new one if none exists.

	w = PushButton(
		parent=None,
		setObjectName='b000',
		setText='<hl style="color:black;">A QPushButton <hl style="color:violet;"><b>with Rich Text</b></hl>',
		resize=QSize(125, 45),
		setWhatsThis='',
		# setVisible=True,
	)

	w.show()
	sys.exit(app.exec_())



# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

'''
Promoting a widget in designer to use a custom class:
>	In Qt Designer, select all the widgets you want to replace, 
		then right-click them and select 'Promote to...'. 

>	In the dialog:
		Base Class:		Class from which you inherit. ie. QWidget
		Promoted Class:	Name of the class. ie. "MyWidget"
		Header File:	Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>	Then click "Add", "Promote", 
		and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
'''

# Deprecated: --------------------




# !/usr/bin/python
# coding=utf-8
from uitk.slots import Slots



class Crease(Slots):
	'''
	'''
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		'''
		'''
		dh = self.sb.crease.draggable_header
		dh.ctxMenu.add(self.sb.ComboBox, setObjectName='cmb000', setToolTip='Crease Editors')

		tb000 = self.sb.crease.tb000
		tb000.ctxMenu.add('QSpinBox', setPrefix='Crease Amount: ', setObjectName='s003', setMinMax_='0-10 step1', setValue=10, setToolTip='Crease amount 0-10. Overriden if "max" checked.')
		tb000.ctxMenu.add('QCheckBox', setText='Toggle Max', setObjectName='chk003', setToolTip='Toggle crease amount from it\'s current value to the maximum amount.')
		tb000.ctxMenu.add('QCheckBox', setText='Un-Crease', setObjectName='chk002', setToolTip='Un-crease selected components or If in object mode, uncrease all.')
		tb000.ctxMenu.add('QCheckBox', setText='Perform Normal Edge Hardness', setObjectName='chk005', setChecked=True, setToolTip='Toggle perform normal edge hardness.')
		tb000.ctxMenu.add('QSpinBox', setPrefix='Hardness Angle: ', setObjectName='s004', setMinMax_='0-180 step1', setValue=30, setToolTip='Normal edge hardness 0-180.')
		tb000.ctxMenu.add('QCheckBox', setText='Crease Vertex Points', setObjectName='chk004', setChecked=True, setToolTip='Crease vertex points.')
		tb000.ctxMenu.add('QCheckBox', setText='Auto Crease', setObjectName='chk011', setToolTip='Auto crease selected object(s) within the set angle tolerance.')
		tb000.ctxMenu.add('QSpinBox', setPrefix='Auto Crease: Low: ', setObjectName='s005', setMinMax_='0-180 step1', setValue=85, setToolTip='Auto crease: low angle constraint.')
		tb000.ctxMenu.add('QSpinBox', setPrefix='Auto Crease: high: ', setObjectName='s006', setMinMax_='0-180 step1', setValue=95, setToolTip='Auto crease: max angle constraint.')
		self.sb.toggleWidgets(tb000.ctxMenu, setDisabled='s005,s006')


	def s003(self, value=None):
		'''Crease Amount
		Tracks the standard crease amount while toggles such as un-crease, and crease max temporarily change the spinbox value. 
		'''
		if not self.sb.crease.tb000.ctxMenu.chk002.isChecked(): #un-crease
			if not self.sb.crease.tb000.ctxMenu.chk003.isChecked(): #toggle max
				self.creaseValue = self.sb.crease.tb000.ctxMenu.s003.value()
				text = self.sb.crease.tb000.text().split()[0]
				self.sb.crease.tb000.setText('{} {}'.format(text, self.creaseValue))


	def draggable_header(self, state=None):
		'''Context menu
		'''
		dh = self.sb.crease.draggable_header


	def chk002(self, state=None):
		'''Un-Crease
		'''
		if self.sb.crease.tb000.ctxMenu.chk002.isChecked():
			self.sb.crease.tb000.ctxMenu.s003.setValue(0) #crease value
			self.sb.crease.tb000.ctxMenu.s004.setValue(180) #normal angle
			self.sb.toggleWidgets(self.sb.crease.tb000.ctxMenu, self.crease_submenu.tb000.ctxMenu, setChecked='chk002', setUnChecked='chk003')
			self.sb.crease.tb000.ctxMenu.s003.setDisabled(True)
			text = 'Un-Crease 0'
		else:
			self.sb.crease.tb000.ctxMenu.s003.setValue(self.creaseValue) #crease value
			self.sb.crease.tb000.ctxMenu.s004.setValue(30) #normal angle
			self.sb.crease.tb000.ctxMenu.s003.setEnabled(True)
			text = '{} {}'.format('Crease', self.creaseValue)

		self.sb.crease.tb000.setText(text)


	def chk003(self, state=None):
		'''Crease: Max
		'''
		if self.sb.crease.tb000.ctxMenu.chk003.isChecked():
			self.sb.crease.tb000.ctxMenu.s003.setValue(10) #crease value
			self.sb.crease.tb000.ctxMenu.s004.setValue(30) #normal angle
			self.sb.toggleWidgets(self.sb.crease.tb000.ctxMenu, self.crease_submenu.tb000.ctxMenu, setChecked='chk003', setUnChecked='chk002')
			self.sb.crease.tb000.ctxMenu.s003.setDisabled(True)
			text = 'Un-Crease 0'
		else:
			self.sb.crease.tb000.ctxMenu.s003.setValue(self.creaseValue) #crease value
			self.sb.crease.tb000.ctxMenu.s004.setValue(60) #normal angle
			self.sb.crease.tb000.ctxMenu.s003.setEnabled(True)
			text = '{} {}'.format('Crease', self.creaseValue)

		self.sb.crease.tb000.setText(text)


	def chk011(self, state=None):
		'''Crease: Auto
		'''
		if self.sb.crease.tb000.ctxMenu.chk011.isChecked():
			self.sb.toggleWidgets(self.sb.crease.tb000.ctxMenu, setEnabled='s005,s006')
		else:
			self.sb.toggleWidgets(self.sb.crease.tb000.ctxMenu, setDisabled='s005,s006')


	def b000(self):
		'''Crease Set Transfer: Transform Node
		'''
		if self.sb.crease.b001.isChecked():
			newObject = str(pm.ls(selection=1)) #ex. [nt.Transform(u'pSphere1')]

			index1 = newObject.find("u")
			index2 = newObject.find(")")
			newObject = newObject[index1+1:index2].strip("'") #ex. pSphere1

			if newObject != "[":
				self.sb.crease.b001.setText(newObject)
			else:
				self.sb.crease.b001.setText("must select obj first")
				self.sb.toggleWidgets(setUnChecked='b001')
			if self.sb.crease.b000.isChecked():
				self.sb.toggleWidgets(setEnabled='b052')
		else:
			self.sb.crease.b001.setText("Object")


	def b001(self):
		'''Crease Set Transfer: Crease Set
		'''
		if self.sb.crease.b000.isChecked():
			creaseSet = str(pm.ls(selection=1)) #ex. [nt.CreaseSet(u'creaseSet1')]

			index1 = creaseSet.find("u")
			index2 = creaseSet.find(")")
			creaseSet = creaseSet[index1+1:index2].strip("'") #ex. creaseSet1

			if creaseSet != "[":
				self.sb.crease.b000.setText(creaseSet)
			else:
				self.sb.crease.b000.setText("must select set first")
				self.sb.toggleWidgets(setUnChecked='b000')
			if self.sb.crease.b001.isChecked():
				self.sb.toggleWidgets(setEnabled='b052')
		else:
			self.sb.crease.b000.setText("Crease Set")
	


# !/usr/bin/python
# coding=utf-8
from uitk.slots.blender import *
from uitk.slots.convert import Convert



class Convert_blender(Convert, Slots_blender):
	def __init__(self, *args, **kwargs):
		Slots_blender.__init__(self, *args, **kwargs)
		Convert.__init__(self, *args, **kwargs)

		cmb = self.sb.convert.draggable_header.ctxMenu.cmb000
		items = ['']
		cmb.addItems_(items, '')

		cmb = self.sb.convert.cmb001
		items = [] 
		contents = cmb.addItems_(items, 'Convert To')


	def cmb000(self, index=-1):
		'''Editors
		'''
		cmb = self.sb.convert.draggable_header.ctxMenu.cmb000

		if index>0:
			text = cmb.items[index]
			if text=='':
				pass
			cmb.setCurrentIndex(0)


	def cmb001(self, index=-1):
		'''Convert To
		'''
		cmb = self.sb.convert.cmb001

		if index>0:
			text = cmb.items[index]
			if text=='NURBS to Polygons': #index 1
				mel.eval('performnurbsToPoly 0;')
			elif text=='NURBS to Subdiv': #index 2
				mel.eval('performSubdivCreate 0;')
			elif text=='Polygons to Subdiv': #index 3
				mel.eval('performSubdivCreate 0;')
			elif text=='Smooth Mesh Preview to Polygons': #index 4
				mel.eval('performSmoothMeshPreviewToPolygon;')
			elif text=='Polygon Edges to Curve': #index 5
				mel.eval('polyToCurve -form 2 -degree 3 -conformToSmoothMeshPreview 1;')
			elif text=='Type to Curves': #index 6
				mel.eval('convertTypeCapsToCurves;')
			elif text=='Subdiv to Polygons': #index 7
				mel.eval('performSubdivTessellate  false;')
			elif text=='Subdiv to NURBS': #index 8
				mel.eval('performSubdToNurbs 0;')
			elif text=='NURBS Curve to Bezier': #index 9
				mel.eval('nurbsCurveToBezier;')
			elif text=='Bezier Curve to NURBS': #index 10
				mel.eval('bezierCurveToNurbs;')
			elif text=='Paint Effects to NURBS': #index 11
				mel.eval('performPaintEffectsToNurbs  false;')
			elif text=='Paint Effects to Curves': #index 12
				mel.eval('performPaintEffectsToCurve  false;')
			elif text=='Texture to Geometry': #index 13
				mel.eval('performTextureToGeom 0;')
			elif text=='Displacement to Polygons': #index 14
				mel.eval('displacementToPoly;')
			elif text=='Displacement to Polygons with History': #index 15
				mel.eval('setupAnimatedDisplacement;')
			elif text=='Fluid to Polygons': #index 16
				mel.eval('fluidToPoly;')
			elif text=='nParticle to Polygons': #index 17
				mel.eval('particleToPoly;')
			elif text=='Instance to Object': #index 18
				mel.eval('convertInstanceToObject;')
			elif text=='Geometry to Bounding Box': #index 19
				mel.eval('performGeomToBBox 0;')
			elif text=='Convert XGen Primitives to Polygons': #index 20
				import xgenm.xmaya.xgmConvertPrimToPolygon as cpp
				cpp.convertPrimToPolygon(False)

			cmb.setCurrentIndex(0)









#module name
print (__name__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# !/usr/bin/python
# coding=utf-8
from uitk.slots import Slots



class Create(Slots):
	'''
	'''
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		'''
		'''
		dh = self.sb.create.draggable_header
		dh.ctxMenu.add(self.sb.ComboBox, setObjectName='cmb000', setToolTip='')

		cmb000 = self.sb.create.draggable_header.ctxMenu.cmb000
		items = ['']
		cmb000.addItems_(items, '')

		cmb001 = self.sb.create.cmb001
		items = ['Polygon', 'NURBS', 'Light']
		cmb001.addItems_(items)

		cmb002 = self.sb.create.cmb002
		items = ["Cube", "Sphere", "Cylinder", "Plane", "Circle", "Cone", "Pyramid", "Torus", "Tube", "GeoSphere", "Platonic Solids", "Text"]
		cmb002.addItems_(items)

		tb000 = self.sb.create.tb000
		tb000.ctxMenu.add('QCheckBox', setText='Translate', setObjectName='chk000', setChecked=True, setToolTip='Move the created object to the center point of any selected object(s).')
		tb000.ctxMenu.add('QCheckBox', setText='Scale', setObjectName='chk001', setChecked=True, setToolTip='Uniformly scale the created object to match the averaged scale of any selected object(s).')


	def draggable_header(self, state=None):
		'''Context menu
		'''
		dh = self.sb.create.draggable_header


	def createPrimitive(self, catagory1, catagory2):
		'''Create a primitive object.

		Parameters:
			catagory1 (str): type
			catagory2 (str): type
		Return:
			(obj) node

		Example: createPrimitive('Polygons', 'Cube')
		'''
		cmb001 = self.sb.create.cmb001
		cmb002 = self.sb.create.cmb002

		cmb001.setCurrentIndex(cmb001.findText(catagory1))
		cmb002.setCurrentIndex(cmb002.findText(catagory2))
		return self.tb000()


	def b001(self):
		'''Create poly cube
		'''
		self.createPrimitive('Polygon', 'Cube')


	def b002(self):
		'''Create poly sphere
		'''
		self.createPrimitive('Polygon', 'Sphere')


	def b003(self):
		'''Create poly cylinder
		'''
		self.createPrimitive('Polygon', 'Cylinder')


	def b004(self):
		'''Create poly plane
		'''
		self.createPrimitive('Polygon', 'Plane')
	


# !/usr/bin/python
# coding=utf-8
import sys, os.path

import importlib
import inspect


def __getattr__(attr_name):
	"""This function dynamically imports a module and returns an attribute from the module. 

	Parameters:
		attr_name (str): The name of the attribute to be imported. The name should be in the format 
					'module_name.attribute_name' or just 'attribute_name'.
	Return:
		(obj) The attribute specified by the `attr_name` argument.

	:Raises:
		AttributeError: If the specified attribute is not found in the original module.

	Example:
		<package>.__getattr__('module1.attribute1') #returns: <attribute1 value>
		<package>.__getattr__('attribute1') #returns: <attribute1 value>
	"""
	try:
		module = __import__(f"{__package__}.{attr_name}", fromlist=[f"{attr_name}"])
		setattr(sys.modules[__name__], attr_name, getattr(module, attr_name))
		return getattr(module, attr_name)

	except ModuleNotFoundError as error:
		raise AttributeError(f"Module '{__package__}' has no attribute '{attr_name}'") from error

# --------------------------------------------------------------------------------------------









# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------

'''
EXAMPLE USE CASE:
import ui.widgets as wgts

wgts.PushButton #get a specific widget.
wgts.widgets #get a list of all widgets.
'''

# --------------------------------------------------------------------------------------------
# deprecated:
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
from uitk.slots.max import *
from uitk.slots.nurbs import Nurbs


class Nurbs_max(Nurbs, Slots_max):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		cmb000 = self.sb.nurbs.draggable_header.ctxMenu.cmb000
		items = []
		cmb000.addItems_(items, 'Curve Editors')

		cmb001 = self.sb.nurbs.cmb001
		items = []
		cmb001.addItems_(items, 'Create Curve')


	def cmb000(self, index=-1):
		'''Maya Curve Operations
		'''
		cmb = self.sb.nurbs.draggable_header.ctxMenu.cmb000

		if index>0:
			text = cmb.items[index]
			if text=='Project Curve':
				mel.eval("ProjectCurveOnSurfaceOptions;")
			elif text=='Duplicate Curve':
				mel.eval("DuplicateCurveOptions;")
			elif text=='Create Curve from Poly':
				mel.eval("CreateCurveFromPolyOptions")
			elif text=='Bend Curve':
				mel.eval("BendCurvesOptions;")
			elif text=='Curl Curve':
				mel.eval("CurlCurvesOptions;")
			elif text=='Modify Curve Curvature':
				mel.eval("ScaleCurvatureOptions;")
			elif text=='Smooth Curve':
				mel.eval("SmoothHairCurvesOptions;")
			elif text=='Straighten Curves':
				mel.eval("StraightenCurvesOptions;")
			elif text=='Extrude Curves':
				mel.eval("ExtrudeOptions;")
			elif text=='Revolve Curves':
				mel.eval("RevolveOptions;")
			elif text=='Loft Curves':
				mel.eval("LoftOptions;")
			elif text=='Planar Curves':
				mel.eval("PlanarOptions;")
			elif text=='Insert Isoparms':
				mel.eval("InsertIsoparmsOptions;")
			elif text=='Insert Knot':
				mel.eval("InsertKnotOptions;")
			elif text=='Rebuild Curve':
				mel.eval("RebuildCurveOptions;")
			elif text=='Extend Curve':
				mel.eval("ExtendCurveOptions;")
			elif text=='Extend Curve On Surface':
				mel.eval("ExtendCurveOnSurfaceOptions;")
			cmb.setCurrentIndex(0)


	def cmb001(self, index=-1):
		'''Create: Curve
		'''
		cmb = self.sb.nurbs.cmb001

		if index>0:
			text = cmb.items[index]
			if text=='Ep Curve Tool':
				mel.eval('EPCurveToolOptions;') #mel.eval('EPCurveTool;')
			elif text=='CV Curve Tool':
				mel.eval('CVCurveToolOptions') #mel.eval('CVCurveTool')
			elif text=='Bezier Curve Tool':
				mel.eval('CreateBezierCurveToolOptions') #mel.eval('CreateBezierCurveTool;')
			elif text=='Pencil Curve Tool':
				mel.eval('PencilCurveToolOptions;') #mel.eval('PencilCurveTool;')
			elif text=='2 Point Circular Arc':
				mel.eval('TwoPointArcToolOptions;') #mel.eval("TwoPointArcTool;")
			elif text=='3 Point Circular Arc':
				mel.eval("ThreePointArcToolOptions;") #mel.eval("ThreePointArcTool;")
			cmb.setCurrentIndex(0)


	@Slots_max.attr
	def tb000(self, state=None):
		'''Revolve
		'''
		tb = self.sb.nurbs.tb000

		degree = tb.ctxMenu.s002.value()
		startSweep = tb.ctxMenu.s003.value()
		endSweep = tb.ctxMenu.s004.value()
		sections = tb.ctxMenu.s005.value()
		range_ = tb.ctxMenu.chk006.isChecked()
		polygon = 1 if tb.ctxMenu.chk007.isChecked() else 0
		autoCorrectNormal = tb.ctxMenu.chk008.isChecked()
		useTolerance = tb.ctxMenu.chk009.isChecked()
		tolerance = tb.ctxMenu.s006.value()

		return pm.revolve(curves, po=polygon, rn=range_, ssw=startSweep, esw=endSweep, ut=useTolerance, tol=tolerance, degree=degree, s=sections, ulp=1, ax=[0,1,0])


	@Slots_max.attr
	def tb001(self, state=None):
		'''Loft
		'''
		tb = self.sb.nurbs.tb001

		uniform = tb.ctxMenu.chk000.isChecked()
		close = tb.ctxMenu.chk001.isChecked()
		degree = tb.ctxMenu.s000.value()
		autoReverse = tb.ctxMenu.chk002.isChecked()
		sectionSpans = tb.ctxMenu.s001.value()
		range_ = tb.ctxMenu.chk003.isChecked()
		polygon = 1 if tb.ctxMenu.chk004.isChecked() else 0
		reverseSurfaceNormals = tb.ctxMenu.chk005.isChecked()
		angleLoftBetweenTwoCurves = tb.ctxMenu.chk010.isChecked()
		angleLoftSpans = tb.ctxMenu.s007.value()

		if angleLoftBetweenTwoCurves:
			start, end = pm.ls(sl=1)[:2] #get the first two selected edge loops or curves.
			return Slots_max.angleLoftBetweenTwoCurves(start, end, count=angleLoftSpans, cleanup=True, uniform=uniform, close=close, autoReverse=autoReverse, degree=degree, sectionSpans=sectionSpans, range=range_, polygon=polygon, reverseSurfaceNormals=reverseSurfaceNormals)

		return pm.loft(curves, u=uniform, c=close, ar=autoReverse, d=degree, ss=sectionSpans, rn=range_, po=polygon, rsn=reverseSurfaceNormals)


	def b012(self):
		'''Project Curve
		'''
		mel.eval("projectCurve;") #ProjectCurveOnMesh;


	def b014(self):
		'''Duplicate Curve
		'''
		pm.mel.DuplicateCurve()


	def b016(self):
		'''Extract Curve
		'''
		pm.mel.CreateCurveFromPoly()


	def b018(self):
		'''Lock Curve
		'''
		pm.mel.LockCurveLength()


	def b019(self):
		'''Unlock Curve
		'''
		pm.mel.UnlockCurveLength()


	def b020(self):
		'''Bend Curve
		'''
		pm.mel.BendCurves()


	def b022(self):
		'''Curl Curve
		'''
		pm.mel.CurlCurves()


	def b024(self):
		'''Modify Curve Curvature
		'''
		pm.mel.ScaleCurvature()


	def b026(self):
		'''Smooth Curve
		'''
		pm.mel.SmoothHairCurves()


	def b028(self):
		'''Straighten Curve

		'''
		pm.mel.StraightenCurves()


	def b030(self):
		'''Extrude

		'''
		pm.mel.Extrude()


	def b036(self):
		'''Planar
		'''
		pm.mel.Planar()


	def b038(self):
		'''Insert Isoparm
		'''
		pm.mel.InsertIsoparms()


	def b040(self):
		'''Edit Curve Tool
		'''
		pm.mel.CurveEditTool()


	def b041(self):
		'''Attach Curve
		'''
		pm.mel.AttachCurveOptions()


	def b042(self):
		'''Detach Curve
		'''
		pm.mel.DetachCurve()


	def b043(self):
		'''Extend Curve
		'''
		pm.mel.ExtendCurveOptions()


	def b045(self):
		'''Cut Curve
		'''
		pm.mel.CutCurve()


	def b046(self):
		'''Open/Close Curve
		'''
		pm.mel.OpenCloseCurve()


	def b047(self):
		'''Insert Knot
		'''
		pm.mel.InsertKnot()


	def b049(self):
		'''Add Points Tool
		'''
		pm.mel.AddPointsTool()


	def b051(self):
		'''Reverse Curve

		'''
		mel.eval("reverse;")


	def b052(self):
		'''Extend Curve
		'''
		pm.mel.ExtendCurve()


	def b054(self):
		'''Extend On Surface
		'''
		pm.mel.ExtendCurveOnSurface()









#module name
print (__name__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# !/usr/bin/python
# coding=utf-8
from uitk.slots import Slots



class Cameras(Slots):
	'''
	'''
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		'''
		'''
		# tree = self.sb.cameras_lower_submenu.tree000
		# tree.expandOnHover = True
		# tree.convert(tree.getTopLevelItems(), 'QLabel') #construct the tree using the existing contents.

		



# !/usr/bin/python
# coding=utf-8
from uitk.slots import Slots



class Uv(Slots):
	'''
	'''
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		'''
		'''
		dh = self.sb.uv.draggable_header
		dh.ctxMenu.add(self.sb.ComboBox, setObjectName='cmb000', setToolTip='Maya UV Editors')
		dh.ctxMenu.add('QPushButton', setText='Create UV Snapshot', setObjectName='b001', setToolTip='Save an image file of the current UV layout.')

		cmb001 = self.sb.uv.cmb001
		cmb001.popupStyle = 'qmenu'
		cmb001.menu_.add(self.sb.CheckBox, setObjectName='chk014', setText='Checkered', setToolTip='')
		cmb001.menu_.add(self.sb.CheckBox, setObjectName='chk015', setText='Borders', setToolTip='')
		cmb001.menu_.add(self.sb.CheckBox, setObjectName='chk016', setText='Distortion', setToolTip='')

		cmb003 = self.sb.uv.cmb003
		self.getMapSize = lambda: int(self.sb.uv.cmb003.currentText()) #get the map size from the combobox as an int. ie. 2048

		tb001 = self.sb.uv.tb001
		tb001.ctxMenu.add('QRadioButton', setText='Standard', setObjectName='chk000', setChecked=True, setToolTip='Create UV texture coordinates for the selected object or faces by automatically finding the best UV placement using simultanious projections from multiple planes.')
		tb001.ctxMenu.add('QCheckBox', setText='Scale Mode 1', setObjectName='chk001', setTristate=True, setChecked=True, setToolTip='0 - No scale is applied.<br>1 - Uniform scale to fit in unit square.<br>2 - Non proportional scale to fit in unit square.')
		tb001.ctxMenu.add('QRadioButton', setText='Seam Only', setObjectName='chk002', setToolTip='Cut seams only.')
		tb001.ctxMenu.add('QRadioButton', setText='Planar', setObjectName='chk003', setToolTip='Create UV texture coordinates for the current selection by using a planar projection shape.')
		tb001.ctxMenu.add('QRadioButton', setText='Cylindrical', setObjectName='chk004', setToolTip='Create UV texture coordinates for the current selection, using a cylidrical projection that gets wrapped around the mesh.<br>Best suited for completely enclosed cylidrical shapes with no holes or projections on the surface.')
		tb001.ctxMenu.add('QRadioButton', setText='Spherical', setObjectName='chk005', setToolTip='Create UV texture coordinates for the current selection, using a spherical projection that gets wrapped around the mesh.<br>Best suited for completely enclosed spherical shapes with no holes or projections on the surface.')
		tb001.ctxMenu.add('QRadioButton', setText='Normal-Based', setObjectName='chk006', setToolTip='Create UV texture coordinates for the current selection by creating a planar projection based on the average vector of it\'s face normals.')
		# tb001.ctxMenu.chk001.toggled.connect(lambda state: self.sb.toggleWidgets(tb001.ctxMenu, setUnChecked='chk002-3') if state==1 else None)

		tb003 = self.sb.uv.tb003
		tb003.ctxMenu.add('QRadioButton', setText='Back-Facing', setObjectName='chk008', setToolTip='Select all back-facing (using counter-clockwise winding order) components for the current selection.')
		tb003.ctxMenu.add('QRadioButton', setText='Front-Facing', setObjectName='chk009', setToolTip='Select all front-facing (using counter-clockwise winding order) components for the current selection.')
		tb003.ctxMenu.add('QRadioButton', setText='Overlapping', setObjectName='chk010', setToolTip='Select all components that share the same uv space.')
		tb003.ctxMenu.add('QRadioButton', setText='Non-Overlapping', setObjectName='chk011', setToolTip='Select all components that do not share the same uv space.')
		tb003.ctxMenu.add('QRadioButton', setText='Texture Borders', setObjectName='chk012', setToolTip='Select all components on the borders of uv shells.')
		tb003.ctxMenu.add('QRadioButton', setText='Unmapped', setObjectName='chk013', setChecked=True, setToolTip='Select unmapped faces in the current uv set.')

		tb004 = self.sb.uv.tb004
		tb004.ctxMenu.add('QCheckBox', setText='Optimize', setObjectName='chk017', setChecked=True, setToolTip='The Optimize UV Tool evens out the spacing between UVs on a mesh, fixing areas of distortion (overlapping UVs).')
		tb004.ctxMenu.add('QCheckBox', setText='Orient', setObjectName='chk007', setChecked=True, setToolTip='Orient selected UV shells to run parallel with the most adjacent U or V axis.')
		tb004.ctxMenu.add('QCheckBox', setText='Stack Similar', setObjectName='chk022', setChecked=True, setToolTip='Stack only shells that fall within the set tolerance.')
		tb004.ctxMenu.add('QDoubleSpinBox', setPrefix='Tolerance: ', setObjectName='s000', setMinMax_='0.0-10 step.1', setValue=1.0, setToolTip='Stack shells with uv\'s within the given range.')

		tb005 = self.sb.uv.tb005
		tb005.ctxMenu.add('QSpinBox', setPrefix='Angle: ', setObjectName='s001', setMinMax_='0-360 step1', setValue=30, setToolTip='Set the maximum angle used for straightening uv\'s.')
		tb005.ctxMenu.add('QCheckBox', setText='Straighten U', setObjectName='chk018', setChecked=True, setToolTip='Unfold UV\'s along a horizonal contraint.')
		tb005.ctxMenu.add('QCheckBox', setText='Straighten V', setObjectName='chk019', setChecked=True, setToolTip='Unfold UV\'s along a vertical constaint.')
		tb005.ctxMenu.add('QCheckBox', setText='Straighten Shell', setObjectName='chk020', setToolTip='Straighten a UV shell by unfolding UV\'s around a selected UV\'s edgeloop.')

		tb006 = self.sb.uv.tb006
		tb006.ctxMenu.add('QRadioButton', setText='Distribute U', setObjectName='chk023', setChecked=True, setToolTip='Distribute along U.')
		tb006.ctxMenu.add('QRadioButton', setText='Distribute V', setObjectName='chk024', setToolTip='Distribute along V.')

		tb008 = self.sb.uv.tb008
		tb008.ctxMenu.add('QCheckBox', setText='To Similar', setObjectName='chk025', setToolTip='Instead of manually selecting what to transfer to; search the scene for similar objects.')
		tb008.ctxMenu.add('QDoubleSpinBox', setPrefix='Tolerance: ', setObjectName='s013', setMinMax_='0.0-1000 step.5', setValue=0.00, setToolTip='The maximum attribute value tolerance to use when searching for similar objects.')
		tb008.ctxMenu.add('QCheckBox', setText='Delete History', setObjectName='chk026', setChecked=True, setToolTip='Remove construction history for the objects transferring from.\nOtherwise, the UV\'s will be lost should any of the frm objects be deleted.')


	def draggable_header(self, state=None):
		'''Context menu
		'''
		dh = self.sb.uv.draggable_header


	def chk001(self, state):
		'''Auto Unwrap: Scale Mode CheckBox
		'''
		tb = self.sb.uv.tb001
		if state==0:
			tb.ctxMenu.chk001.setText('Scale Mode 0')
		if state==1:
			tb.ctxMenu.chk001.setText('Scale Mode 1')
			self.sb.toggleWidgets(tb.ctxMenu, setUnChecked='chk002-6')
		if state==2:
			tb.ctxMenu.chk001.setText('Scale Mode 2')


	def b021(self):
		'''Unfold and Pack
		'''
		self.tb004() #perform unfold
		self.tb000() #perform pack


	def b022(self):
		'''Cut UV hard edges
		'''
		self.sb.selection.slots.tb003() #perform select edges by angle.
		self.b005() #perform cut.


	def b023(self):
		'''Move To Uv Space: Left
		'''
		self.moveSelectedToUvSpace(-1, 0) #move left


	def b024(self):
		'''Move To Uv Space: Down
		'''
		self.moveSelectedToUvSpace(0, -1) #move down


	def b025(self):
		'''Move To Uv Space: Up
		'''
		self.moveSelectedToUvSpace(0, 1) #move up


	def b026(self):
		'''Move To Uv Space: Right
		'''
		self.moveSelectedToUvSpace(1, 0) #move right


	
